#!/usr/bin/env python
import importlib
import os
import re
import shutil
from datetime import datetime
from typing import Any

import numpy as np  # pyright: ignore[reportMissingImports]
import pandas as pd
import rich.console
from rich.prompt import Prompt

import relecov_tools.utils
import relecov_tools.validate
from relecov_tools.base_module import BaseModule
from relecov_tools.config_json import ConfigJson

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class BioinfoReportLog:
    def __init__(self, log_report=None):
        if not log_report:
            self.report = {"error": {}, "valid": {}, "warning": {}}
        else:
            self.report = log_report

    def update_log_report(self, method_name, status, message):
        """Update the progress log report with the given method name, status, and message.

        Args:
            method_name (str): The name of the method being logged.
            status (str): The status of the log message, can be one of 'valid', 'error', or 'warning'.
            message (str): The message to be logged.

        Returns:
            dict: The updated progress log report.

        Raises:
            ValueError: If an invalid status is provided.
        """
        if status == "valid":
            self.report["valid"].setdefault(method_name, []).append(message)
            return self.report
        elif status == "error":
            self.report["error"].setdefault(method_name, []).append(message)
            return self.report
        elif status == "warning":
            self.report["warning"].setdefault(method_name, []).append(message)
            return self.report
        else:
            raise ValueError("Invalid status provided.")

    def print_log_report(self, name, sections):
        """Prints the log report by calling util's function.

        Args:
            name (str): The name of the log report.
            sections (list of str): The sections of the log report to be printed.

        Returns:
            None
        """
        relecov_tools.utils.print_log_report(self.report, name, sections)


# TODO: Add method to validate bioinfo_config.json file requirements.
class BioinfoMetadata(BaseModule):
    def __init__(
        self,
        json_file: str | None = None,
        json_schema_file: str | None = None,
        input_folder: str | None = None,
        output_dir: str | None = None,
        software_name: str | None = None,
        update: bool = False,
        soft_validation: bool = False,
    ):

        super().__init__(output_dir=output_dir, called_module=__name__)
        self.log.info("Initiating read-bioinfo-metadata process")
        # Init params
        self.update = update
        self.soft_validation = soft_validation

        # Init logs
        self.log_report = BioinfoReportLog()
        self.logsum = self.parent_log_summary(output_dir=output_dir)

        # Init config
        self.config_json = ConfigJson()

        # Init attributes
        self._init_input_folder(input_folder)
        self._init_software_name(software_name)
        self._init_json_schema(json_schema_file)
        self._init_output_dir(output_dir)
        self._init_readlabmeta_json(json_file)
        self._init_j_data_and_batch()

        # Init config
        self._init_bioinfo_config()

    def _init_json_schema(self, json_schema_file: str | None = None) -> None:
        """Initializes the JSON schema for bioinformatics metadata.

        Args:
            json_schema_file (str) optional
                Path to the JSON schema file. If not provided, it will use the default schema defined in the config.

        Returns:
            None
        """
        if json_schema_file is None:
            schema_name = self.config_json.get_topic_data("generic", "relecov_schema")
            json_schema_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schema", str(schema_name)
            )
        self.json_schema = relecov_tools.utils.read_json_file(json_schema_file)

    def _init_output_dir(self, output_dir: str | None = None) -> None:
        """Initializes the output directory for storing results.

        Args:
            output_dir (str | None)
                Path to the output directory. If not provided, it will prompt the user to select one.

        Returns:
            None
        """
        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = os.path.realpath(output_dir)

    def _init_readlabmeta_json(self, json_file: str | None = None) -> None:
        """Initializes the readlab metadata JSON file.

        Args:
            json_file (str | None): Path to the readlab metadata JSON file.
            If not provided, it will prompt the user to select one.

        Raises:
            ValueError
                If the provided JSON file does not exist or is not a valid file.
        Returns:
            None
        """
        if json_file is None:
            json_file = relecov_tools.utils.prompt_path(
                msg="Select the json file that was created by the pipeline-manager or read-lab-metadata module."
            )
        if json_file is None or not os.path.isfile(json_file):
            self.update_all_logs(
                self.__init__.__name__,
                "error",
                f"file {json_file} does not exist",
            )
            self.log_report.print_log_report(self.__init__.__name__, ["error"])
            raise ValueError(f"Json file {json_file} does not exist, cannot continue.")
        self.readlabmeta_json_file = json_file

    def _init_j_data_and_batch(self) -> None:
        """Initializes the j_data and batch ID from the readlab metadata JSON file.
        Args:
            None
        Returns:
            None
        """
        stderr.print("[blue]Reading lab metadata json")
        self.j_data = self.collect_info_from_lab_json()
        batch_id = self.get_batch_id_from_data(self.j_data)
        self.set_batch_id(batch_id)

    def _init_input_folder(self, input_folder: str | None = None) -> None:
        """Initializes the input folder for bioinformatics analysis files.

        Args:
        input_folder (str | None)
            Path to the input folder. If not provided, it will prompt the user to select one.
        Returns:
            None
        """
        if input_folder is None:
            self.input_folder = relecov_tools.utils.prompt_path(
                msg="Select the input folder"
            )
        else:
            self.input_folder = input_folder

    def _init_software_name(self, software_name: str | None = None) -> None:
        """Initializes the software name used in the bioinformatic analysis.

        Args:
        software_name (str | None)
            Name of the software, pipeline or tool used in the bioinformatic analysis.
            If not provided, it will prompt the user to select one.

        Raises:
        ValueError
            If no software name is provided, it raises a ValueError and logs an error message.

        Returns:
            None
        """
        if software_name is None:
            software_name = relecov_tools.utils.prompt_path(
                msg="Select the software, pipeline or tool use in the bioinformatic analysis: "
            )
        if software_name:
            self.software_name = software_name
        else:
            self.update_all_logs(
                self.__init__.__name__,
                "error",
                "No software name provided, cannot continue.",
            )
            self.log_report.print_log_report(self.__init__.__name__, ["error"])
            raise ValueError("No software name provided, cannot continue.")

    def _init_bioinfo_config(self) -> None:
        """Initializes the bioinformatics configuration based on the provided software name.

        Raises:
            ValueError: If the software name is not found or no configuration is available for it.
        """
        self.bioinfo_json_file = os.path.join(
            os.path.dirname(__file__), "conf", "bioinfo_config.json"
        )

        bioinfo_config = ConfigJson(self.bioinfo_json_file)
        available_software = relecov_tools.utils.get_available_software(
            self.bioinfo_json_file
        )
        method_name = self.__init__.__name__

        # Raises ValueError if the software name is not in the available software list
        if self.software_name not in available_software:
            options = ", ".join(available_software)
            error_msg = (
                f"No configuration available for '{self.software_name}'. "
                f"Currently, the only available software options are: {options}"
            )
            self.update_all_logs(method_name, "error", error_msg)
            self.log_report.print_log_report(self.__init__.__name__, ["error"])
            raise ValueError(error_msg)

        config = bioinfo_config.get_configuration(self.software_name)
        # Raises ValueError if the configuration for the software is None
        if config is None:
            error_msg = (
                f"No configuration found for '{self.software_name}' "
                f"in {self.bioinfo_json_file}."
            )
            self.update_all_logs(method_name, "error", error_msg)
            self.log_report.print_log_report(self.__init__.__name__, ["error"])
            raise ValueError(error_msg)

        # set software configuration indicating that it is a dictionary
        # and access its keys and values directly making sure that's it not going to ever be None.
        self.software_config: dict[str, Any] = config

    def update_all_logs(self, method_name: str, status: str, message: str) -> dict:
        """
        Updates the log report with the given method name, status, and message.

        Args:
        method_name (str)
            The name of the method being logged.
        status (str)
            The status of the log message, can be one of 'valid', 'error', or 'warning'.
        message (str)
            The message to be logged.

        Returns:
            report (dict): The updated log report.
        """
        report = self.log_report.update_log_report(method_name, status, message)
        if status == "error":
            self.logsum.add_error(key=method_name, entry=message)
        elif status == "warning":
            self.logsum.add_warning(key=method_name, entry=message)
        return report

    def scan_directory(
        self,
    ) -> dict[str, list[str]]:
        """Scanns bioinfo analysis directory and identifies files according to the file name patterns defined in the software configuration json.

        Returns:
            files_found (dict): A dictionary containing file paths found based on the definitions provided
            in the bioinformatic JSON file within the software scope (self.software_config).
        """
        # set method name
        method_name = f"{self.scan_directory.__name__}"

        # Get the total number of files and scanned files in all folders in self.input_folder
        total_files, scanned_files_per_folder = self._walk_input_folder()
        files_found = self._find_matching_files(method_name, scanned_files_per_folder)
        # search for files matching the patterns defined in the software configuration
        if files_found:
            # If files are found, update the log report and return the files_found dictionary
            scan_success_msg = f"Scanning process succeed. Scanned {total_files} files."
            self.update_all_logs(
                self.scan_directory.__name__,
                "valid",
                scan_success_msg,
            )
            self.log_report.print_log_report(method_name, ["valid", "warning"])
            return files_found
        else:
            # If no files are found matching the patterns, log an error and raise a ValueError
            err_msg = f"No files found in '{self.input_folder}' according to '{os.path.basename(self.bioinfo_json_file)}' file name patterns."
            self.update_all_logs(
                method_name,
                "error",
                err_msg,
            )
            self.log_report.print_log_report(method_name, ["error"])
            raise ValueError(err_msg)

    def _walk_input_folder(self) -> tuple[int, list[tuple[str, list[str]]]]:
        """Walk through the input folder and collect all files.
        Returns:
        """
        total_files = 0
        scanned_files_per_folder = []
        for root, dirs, files in os.walk(self.input_folder, topdown=True):
            # Skipping nextflow's work junk directory
            # dirs[:] is used to modify the list of directories in place
            dirs[:] = [d for d in dirs if d != "work"]
            total_files += len(files)
            scanned_files_per_folder.append((root, files))
        return total_files, scanned_files_per_folder

    def _find_matching_files(
        self, method_name: str, scanned_files_per_folder: list[tuple[str, list[str]]]
    ) -> dict[str, list[str]]:
        """For each topic in the software configuration, search for matching files.

        Args:
        scanned_files_per_folder (list[tuple[str, list[str]]]): A list of tuples containing the root path and a list of files in that path.
        method_name (str): The name of the method being logged.

        Returns:
        files_found (dict): A dictionary containing file paths found based on the definitions provided in the bioinformatic JSON file within the software scope (self.software_config).
        """
        files_found = {}
        # Iterate over each topic in the software configuration
        for topic_key, topic_scope in self.software_config.items():
            # if topic has not fn (file pattern) defined, skip it
            if not self._topic_has_file_pattern(topic_key, topic_scope, method_name):
                continue

            # get file pattern from topic scope now that we know it has a valid pattern
            file_pattern = topic_scope["fn"]

            # Search for files matching the pattern in the scanned files
            for root_path, file_list in scanned_files_per_folder:
                if matching_files := self._get_matching_files_for_pattern(
                    file_list, root_path, file_pattern
                ):
                    # If matching files are found, add them to the files_found dictionary
                    # set default to an empty list if the topic_key is not already present
                    files_found.setdefault(topic_key, []).extend(matching_files)

        return files_found

    def _topic_has_file_pattern(
        self, topic_key: str, topic_scope: dict, method_name: str
    ) -> bool:
        """Check if topic_scope has a valid pattern; log warning if not.

        Args:
            topic_key (str): The key of the topic in the software configuration.
            topic_scope (dict): The scope of the topic in the software configuration.
            method_name (str): The name of the method being logged.

        Returns:
            bool: True if the topic has a valid file pattern, False otherwise.
        """
        # set pattern
        pattern = topic_scope.get("fn")

        # Check if pattern is valid
        is_invalid_type = not isinstance(pattern, str)
        # Check if pattern is an empty string
        is_empty_string = isinstance(pattern, str) and not pattern.strip()
        # If pattern is invalid or empty, log a warning and return False
        if is_invalid_type or is_empty_string:
            topic_name = f"{self.software_name}.{topic_key}"
            error_msg = f"Invalid or missing 'fn' (file pattern) in '{topic_name}'."
            self.update_all_logs(method_name, "warning", error_msg)
            return False
        return True

    def _get_matching_files_for_pattern(
        self, file_list: list[str], root_path: str, file_pattern: str
    ) -> list[str]:
        """Return a list of files in file_list under root_path that match file_pattern.
        Args:
            file_list (list[str]): A list of file names to search for.
            root_path (str): The root path where the files are located.
            file_pattern (str): The regex pattern to match the files.
        Returns:
            matching_files (list[str]): A list of file paths that match the file_pattern.
        """
        matching_files = []
        # Iterate over each file in the file_list
        for file_name in file_list:
            file_path = os.path.join(root_path, file_name)
            # Check if the file matches the pattern
            # update matching_files with the file matching the pattern
            if re.search(file_pattern, file_path):
                matching_files.append(file_path)
        return matching_files

    def mapping_over_table(
        self,
        j_data: list[dict],
        map_data: dict,
        mapping_fields: dict,
        table_name: list[str],
    ) -> list[dict]:
        """
        Maps bioinformatics metadata from map_data to j_data based on the mapping_fields.

        Args:
            j_data (list[dict]): Metadata from the lab (one item per sample).
            map_data (dict): Processed bioinformatic metadata, indexed by sample name.
            mapping_fields (dict): Mapping from JSON fields to software fields as defined in the config.
            table_name (list[str]): Name or path of the mapping table.

        Returns:
            list[dict]: The updated j_data list with mapped bioinformatic fields.
        """
        # Set method name and initialize error lists
        method_name = f"{self.mapping_over_table.__name__}:{self.software_name}.{self.current_config_key}"
        errors: list[str] = []
        field_errors: dict = {}
        field_valid: dict = {}

        # Get sample ids from j_data and map_data
        sample_ids = {
            sample_name for row in j_data if (sample_name := self._get_sample_name(row))
        }
        # check if samples from j_data match the samples in map_data
        matched_samples = sample_ids & set(map_data.keys())

        # iterate over j_data
        for row in j_data:
            # get sample name from row
            sample_name = self._get_sample_name(row)

            if not sample_name:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"Sequencing_sample_id missing in {row.get('collecting_sample_id')}... Skipping...",
                )
                continue
            # If sample_name is in map_data, map the fields
            if sample_name in map_data:
                self._map_fields_to_row(
                    row,
                    map_data,
                    mapping_fields,
                    sample_name,
                    field_valid,
                    field_errors,
                )
            # If the software configuration allows multiple samples and no samples were matched, handle it accordingly
            # Used for tools that do not have a table with samples but rather individual files
            elif (
                not self.software_config[self.current_config_key].get(
                    "multiple_samples"
                )
                and not matched_samples
            ):
                self._map_non_multiple_sample(
                    row, map_data, mapping_fields, field_valid, field_errors
                )
            else:
                errors.append(sample_name)
                for field in mapping_fields:
                    row[field] = self.config_json.get_topic_data(
                        "generic", "not_provided_field"
                    )

        # work around when map_data comes from several per-sample tables/files instead of single table
        # get the dirname where the tables/files are or basename of the table_name
        resolved_table = (
            os.path.dirname(table_name[0]) if len(table_name) > 2 else table_name[0]
        )

        # Manage errors and field validations
        if errors:
            self.update_all_logs(
                method_name,
                "warning",
                f"{len(errors)} samples missing in '{resolved_table}': {', '.join(errors)}.",
            )
        else:
            self.update_all_logs(
                method_name,
                "valid",
                f"All samples were successfully found in {resolved_table}.",
            )

        if field_errors:
            self.update_all_logs(
                method_name,
                "warning",
                f"Missing fields in {resolved_table}:\n\t{field_errors}",
            )
        else:
            self.update_all_logs(
                method_name,
                "valid",
                f"Successfully mapped fields in {', '.join(field_valid.keys())}.",
            )

        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data

    def _get_sample_name(self, row: dict) -> str | None:
        """
        Generates a sample name from sequencing_sample_id and unique_sample_id if present.

        Args:
            row (dict): The metadata row from j_data.

        Returns:
            str | None: A unique sample identifier, or None if sequencing_sample_id is missing.
        """
        # get sample ids from j_data
        # If unique_sample_id is not present, it will use only sequencing_sample_id.
        # else it will use both to create a unique sample id.
        seq_id = row.get("sequencing_sample_id")
        if not seq_id:
            return None
        uniq_id = row.get("unique_sample_id")
        return f"{seq_id}_{uniq_id}" if uniq_id else seq_id

    def _map_fields_to_row(
        self,
        row: dict,
        map_data: dict,
        mapping_fields: dict,
        sample_name: str,
        field_valid: dict,
        field_errors: dict,
    ) -> None:
        """
        Maps fields for a given sample from map_data using the mapping_fields.

        Args:
            row (dict): The metadata row to update.
            map_data (dict): Dictionary with bioinfo values by sample name.
            mapping_fields (dict): Mapping of fields.
            sample_name (str): The key in map_data.
            field_valid (dict): Accumulator for successfully mapped fields.
            field_errors (dict): Accumulator for mapping errors.
        """
        # Iterate over mapping fields and map values to the row
        for schema_field, table_field in mapping_fields.items():
            try:
                # Get the raw value from map_data for the sample
                raw_val = map_data[sample_name][table_field]
            except KeyError as e:
                field_errors[sample_name] = {schema_field: str(e)}
                row[schema_field] = self.config_json.get_topic_data(
                    "generic", "not_provided_field"
                )
            # Replace NA values if needed
            raw_val = self.replace_na_value_if_needed(schema_field, raw_val)
            # get the expected type from the JSON schema
            expected_type = (
                self.json_schema["properties"]
                .get(schema_field, {})
                .get("type", "string")
            )
            # convert the raw value to the expected type
            row[schema_field] = relecov_tools.utils.cast_value_to_schema_type(
                raw_val, expected_type
            )
            # assign the field to field value validated
            field_valid[sample_name] = {schema_field: table_field}

    def _map_non_multiple_sample(
        self,
        row: dict,
        map_data: dict,
        mapping_fields: dict,
        field_valid: dict,
        field_errors: dict,
    ) -> None:
        """
        Handles mapping for tools where samples are not in a table but individually defined.

        Args:
            row (dict): The metadata row to update.
            map_data (dict): Bioinfo metadata from config-defined keys.
            mapping_fields (dict): Dictionary with nested mappings.
            field_valid (dict): Collector for valid mappings.
            field_errors (dict): Collector for failed mappings.
        """
        # Iterate over mapping fields and map values to the row
        for field, value_dict in mapping_fields.items():
            # iterate over the value_dict which contains json_field and software_key
            for json_field, software_key in value_dict.items():
                try:
                    raw_val = map_data[software_key][field]
                    expected_type = (
                        self.json_schema["properties"]
                        .get(json_field, {})
                        .get("type", "string")
                    )
                    row[json_field] = relecov_tools.utils.cast_value_to_schema_type(
                        raw_val, expected_type
                    )
                    field_valid[software_key] = {json_field: field}
                except KeyError as e:
                    field_errors[software_key] = {json_field: str(e)}
                    row[json_field] = self.config_json.get_topic_data(
                        "generic", "not_provided_field"
                    )

    def validate_sample_names(self) -> None:
        """Validate that the sequencing_sample_id from the JSON input is present in the samples_id.txt.

        Raises:
            ValueError: If no sample from the JSON input matches the samples in the samples_id.txt.
        """
        # Check if samples_id.txt exists
        samples_id_path = os.path.join(self.input_folder, "samples_id.txt")
        with open(samples_id_path, "r") as file:
            samples_id_list = [line.strip() for line in file.readlines()]

        # Get sample names from JSON input.
        # if unique_sample_id is not present, it will use only sequencing_sample_id.
        json_samples = [
            sample_name
            for sample in self.j_data
            if (sample_name := self._get_sample_name(sample)) is not None
        ]
        matching_samples = set(json_samples).intersection(samples_id_list)

        if not matching_samples:
            raise ValueError(
                "No sample from the JSON input matches the samples in the provided analysis folder."
            )
        print(
            f"Found {len(matching_samples)}/{len(json_samples)} matching samples in the samplesheet."
        )
        self.log.info(
            f"Found {len(matching_samples)}/{len(json_samples)} matching samples in the samplesheet."
        )

    def validate_software_mandatory_files(self, files_dict: dict) -> None:
        """Validates the presence of all mandatory files as defined in the software configuration JSON.

        Args:
            files_dict (dict{str:str}): A dictionary containing file paths found based on the definitions provided in the bioinformatic JSON file within the software scope (self.software_config).
        """
        method_name = f"{self.validate_software_mandatory_files.__name__}"
        missing_required = []
        # Iterate over the software configuration keys to check for required files
        for key in self.software_config:
            if key == "fixed_values":
                continue
            if self.software_config[key].get("required") is True:
                try:
                    files_dict[key]
                except KeyError:
                    missing_required.append(key)
                    continue
            else:
                continue
        # If any required files are missing, log an error and raise a ValueError
        if missing_required:
            error_msg = f"Missing mandatory files in {self.software_name}:{', '.join(missing_required)}"
            self.update_all_logs(
                method_name,
                "error",
                error_msg,
            )
            self.log_report.print_log_report(method_name, ["error"])
            raise ValueError(error_msg)
        else:
            self.update_all_logs(
                method_name, "valid", "Successful validation of mandatory files."
            )
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return

    def add_bioinfo_results_metadata(
        self,
        files_dict: dict,
        file_tag: str,
        j_data: list[dict],
        output_dir: str | None = None,
    ) -> tuple[list[dict], list[dict]]:
        """
        Adds metadata from bioinformatics results to j_data.
        Calls file handler and maps handled data into j_data.

        Args:
            files_dict (dict): Mapping of config keys to file paths.
            file_tag (str): Tag used for output filenames.
            j_data (list[dict]): List of lab metadata entries.
            output_dir (str, optional): Path to write outputs.

        Returns:
            tuple: (updated j_data, extra_json_data)
        """

        # Initialize method name and extra_json_data
        method_name = self.add_bioinfo_results_metadata.__name__
        extra_json_data: list[dict] = []
        j_data_mapped = j_data

        # Iterate over each key in the software configuration
        for key in self.software_config:
            # skip workflow_summary and fixed_values keys that will be handled later
            if key in ("workflow_summary", "fixed_values"):
                continue

            self.current_config_key = key
            map_method_name = f"{method_name}:{self.software_name}.{key}"

            if key not in files_dict:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"No file path found for '{self.software_name}.{key}'",
                )
                continue

            stderr.print(f"[blue]Start processing {self.software_name}.{key}")
            self.log.info(f"Start processing {self.software_name}.{key}")

            # Process the config key to get data to map and extra data if any
            data_to_map, extra_data = self._process_config_key(
                key, files_dict[key], file_tag, output_dir, metadata=j_data
            )

            if extra_data:
                extra_json_data.append(extra_data)

            # get mapping fields from the software configuration
            mapping_fields = self.software_config[key].get("content")
            if not mapping_fields:
                self.update_all_logs(
                    map_method_name,
                    "warning",
                    f"No metadata found to perform mapping from '{self.software_name}.{key}' despite 'content' fields being defined.",
                )
                self.log_report.print_log_report(map_method_name, ["warning"])
                continue

            # If data_to_map is not empty, perform mapping
            if data_to_map:
                j_data_mapped = self.mapping_over_table(
                    j_data=j_data,
                    map_data=data_to_map,
                    mapping_fields=mapping_fields,
                    table_name=files_dict[key],
                )
            else:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"No metadata found to perform standard mapping when processing '{self.software_name}.{key}'",
                )

        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data_mapped, extra_json_data

    def _process_config_key(
        self,
        key: str,
        file_path: list[str],
        file_tag: str,
        output_dir: str | None,
        metadata: list[dict] | None = None,
    ) -> tuple[dict | None, dict | None]:
        """
        Handles the data processing logic for a single config key.

        Args:
            key (str): The config key to process.
            file_path (list): Path to the corresponding file.
            file_tag (str): Tag for generated outputs.
            output_dir (str | None): Output directory.
            method_name (str): Name for logging context.

        Returns:
            tuple: (data_to_map, extra_json_data)
        """
        config = self.software_config[key]
        file_name = config.get("fn")
        func_name = config.get("function")
        # If func_name is None, it means we will handle the file as a table with a default function
        if func_name is None:
            data = self.extract_sample_metadata_from_table(file_path, file_name)
        # If func_name is defined, we will process the file using the function defined in the config
        # and present in assets.pipeline_utils
        else:
            extra_kwargs = {}
            if func_name == "utils/parse_long_table":
                extra_kwargs["metadata"] = metadata
            data = self.process_metadata(
                file_path,
                file_tag=file_tag,
                func_name=func_name,
                out_path=output_dir,
                **extra_kwargs,
            )

        if config.get("split_by_batch") and config.get("extra_dict"):
            return None, data
        return data, None

    def extract_sample_metadata_from_table(
        self, file_list: list, conf_tab_name: str
    ) -> dict:
        """Reads a tabular file in different formats and returns a dictionary containing
        the corresponding data for each sample.

        Args:
            file_list (list): A list of file path/s to be processed.
            conf_tab_name (str): Name of the table in the defined config file.

        Returns:
            data (dict): A dictionary containing metadata as defined in handling_files.
        """
        method_name = f"{self.add_bioinfo_results_metadata.__name__}:{self.extract_sample_metadata_from_table.__name__}"
        # get file extension and sample index column position
        file_ext = os.path.splitext(conf_tab_name)[1]
        sample_idx_col_pos = self.get_sample_idx_col_pos(self.current_config_key)
        # allowed file extensions and their separators
        ext_dict = {".csv": ",", ".tsv": "\t", ".tab": "\t"}

        mapping_fields = self.software_config[self.current_config_key].get("content")

        if not mapping_fields:
            return {}

        if conf_tab_name.endswith(".gz"):
            inner_ext = os.path.splitext(conf_tab_name.strip(".gz"))[1]
            if inner_ext in ext_dict:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"Expected tabular file '{conf_tab_name}' is compressed and cannot be processed.",
                )
            return {}

        if file_ext in ext_dict:
            try:
                return relecov_tools.utils.read_csv_file_return_dict(
                    file_name=file_list[0],
                    sep=ext_dict[file_ext],
                    key_position=sample_idx_col_pos,
                )
            except FileNotFoundError as e:
                self.update_all_logs(
                    method_name,
                    "error",
                    f"Tabular file not found: '{file_list[0]}': {e}",
                )
                raise FileNotFoundError(
                    f"Tabular file not found: '{file_list[0]}'"
                ) from e
        else:
            self.update_all_logs(
                method_name,
                "error",
                f"Unrecognized defined file name extension '{file_ext}' in '{conf_tab_name}'.",
            )
            raise ValueError(self.log_report.print_log_report(method_name, ["error"]))

    def process_metadata(
        self,
        file_list: list,
        file_tag: str,
        func_name: str,
        out_path: str | None = None,
        **kwargs,
    ) -> dict:
        """This method dynamically loads and executes the functions specified in config file.
        It is used to apply standard or custom metadata processing depending on the current
        software context (`self.software_name`).

        Args:
            file_list (list): A list of file path/s to be processed.
            func_name (str): The name of the function to execute
            file_tag (str): Tag that will be used for output filenames includes batch date (same as download date) and hex.
            out_path (str): Path to save output files generated during handling_files() process.

        Returns:
            data: A dictionary containing bioinfo metadata handled for each sample.
        """
        method_name = f"{self.add_bioinfo_results_metadata.__name__}:{self.process_metadata.__name__}"
        try:
            # Dynamically import the function from the specified module
            if func_name.startswith("utils/"):
                utils_name = "relecov_tools.assets.pipeline_utils.utils"
                func_name = func_name.split("/", 1)[1]
            else:
                utils_name = f"relecov_tools.assets.pipeline_utils.{self.software_name}"

            # Dynamically import the function from the specified module
            module = importlib.import_module(utils_name)
            # Get method from func_name and execute it.
            func_obj = getattr(module, func_name)
            data = func_obj(
                file_list,
                file_tag=file_tag,
                pipeline_name=self.software_name,
                output_folder=out_path,
                **kwargs,
            )

        except Exception as e:
            self.update_all_logs(
                method_name,
                "error",
                f"Error occurred while parsing '{func_name}': {e}.",
            )
            self.log_report.print_log_report(method_name, ["error"])
            raise ValueError(f"Error occurred while parsing '{func_name}': {e}.") from e
        return data

    def add_fixed_values(self, j_data: list[dict], out_filename: str) -> list[dict]:
        """Add fixed values to j_data as defined in the bioinformatics configuration (definition: "fixed values")

        Args:
            j_data (list(dict{str:str}): A list of dictionaries containing metadata lab (one item per sample).
            out_filename (str): File name of the bioinfo_lab_metadata json

        Returns:
            j_data: updated j_data with fixed values added in it.
        """
        method_name = f"{self.add_fixed_values.__name__}"
        try:
            f_values = self.software_config["fixed_values"]
            for row in j_data:
                row["bioinfo_metadata_file"] = out_filename
                for field, value in f_values.items():
                    row[field] = value
            self.update_all_logs(method_name, "valid", "Fields added successfully.")
        except KeyError as e:
            self.update_all_logs(
                method_name, "warning", f"Error found while adding fixed values: {e}"
            )

        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data

    def replace_na_value_if_needed(self, field: str, raw_val: str | None) -> str | None:
        """
        Replace 'NA', None or NaN with 'Not Provided [SNOMED:434941000124101]'
        if the field is not required in the schema.

        Args:
            field (str): The field name to check.
            raw_val (str | None): The value to check and potentially replace.
        Returns:
            str: The original value if it is not 'NA', None or NaN, otherwise Not Provided.
        """
        required_fields = self.json_schema.get("required", [])
        is_na = (
            raw_val is None
            or (isinstance(raw_val, str) and raw_val.strip().upper() in ["NA", "NONE"])
            or (isinstance(raw_val, float) and np.isnan(raw_val))
        )
        if is_na and field not in required_fields:
            return self.config_json.get_topic_data("generic", "not_provided_field")
        return raw_val

    def map_and_extract_bioinfo_paths(
        self, files_found_dict: dict, j_data: list[dict]
    ) -> list[dict]:
        """
        Adds file paths to j_data for each sample, using the provided files_found_dict
        and the current software configuration.

        Args:
            files_found_dict (dict): Mapping of config keys to file paths found.
            j_data (list[dict]): Sample metadata.

        Returns:
            list[dict]: Updated j_data with file paths.
        """
        method_name = self.map_and_extract_bioinfo_paths.__name__
        sample_name_error = 0
        multiple_sample_files = self.get_multiple_sample_files()

        for row in j_data:
            # ger sample_name that will be a combination of sequencing_sample_id and unique_sample_id
            # if unique_sample_id is not present, it will use only sequencing_sample_id.
            sample_name = self._get_sample_name(row)
            base_cod_path = row.get("sequence_file_path_R1")

            if not sample_name:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"Sequencing_sample_id missing in {row.get('collecting_sample_id')}... Skipping...",
                )
                continue

            if not base_cod_path:
                self.update_all_logs(
                    method_name,
                    "error",
                    f"No 'sequence_file_path_R1' found for sample {sample_name}. Unable to generate paths.",
                )
                continue
            # Iterate over each key (mapping_stats, quality_control, etc.) in files_found_dict
            # Process files associated with each key
            for key, files in files_found_dict.items():
                # select matching file paths based on sample_name and key
                file_paths = self._select_matching_paths(
                    files, sample_name, key, multiple_sample_files
                )
                # set matched file paths to appropriate field in json row
                path_key = self._assign_file_paths_to_row(
                    row, key, file_paths, base_cod_path
                )

                # Extract files to analysis_results folder if configured
                if self.software_config[key].get("extract"):
                    self.extract_file(
                        file=file_paths,
                        dest_folder=base_cod_path,
                        sample_name=sample_name,
                        path_key=path_key,
                    )

        self.log_report.print_log_report(method_name, ["warning"])
        if sample_name_error == 0:
            self.update_all_logs(method_name, "valid", "File paths added successfully.")
        self.log_report.print_log_report(method_name, ["valid", "warning"])

        return j_data

    def _select_matching_paths(
        self, files: list[str], sample_name: str, key: str, multi_sample_keys: list[str]
    ) -> list[str]:
        """
        Selects matching file paths for a given sample name and config key.
        Args:
            files (list[str]): List of file paths to search.
            sample_name (str): The sample name to match against file paths.
            key (str): The configuration key to check against multi-sample keys.
            multi_sample_keys (list[str]): List of keys that allow multiple samples.
        Returns:
            list[str]: Matching paths or placeholder.
        """
        if not files:
            return [self.config_json.get_topic_data("generic", "not_provided_field")]
        # if key is in multi_sample_keys, return all files
        if key in multi_sample_keys:
            return files
        # else return files that match the sample_name
        return [f for f in files if sample_name in f]

    def _assign_file_paths_to_row(
        self,
        row: dict,
        key: str,
        file_paths: list[str],
        base_cod_path: str,
    ) -> None:
        """
        Assigns file path(s) to the row, handling renaming and extraction if needed.
        """
        # get config for the current key
        config = self.software_config[key]
        path_key = config.get("filepath_name", key)

        # get the name of the field configured in config file "filepath_name"
        if config.get("filepath_name"):
            analysis_results_paths = []
            # for each file_path in file_paths
            for f in file_paths:
                # If not provided append and continue
                if f == self.config_json.get_topic_data(
                    "generic", "not_provided_field"
                ):
                    analysis_results_paths.append(f)
                    continue
                # check if configured as extract or function
                extract_or_func = config.get("extract") or config.get("function")
                # if not configured as extract or function, append the file path and continue
                if not extract_or_func:
                    analysis_results_paths.append(f)
                    continue
                # if configured as split_by_batch, rename the file to match the COD
                if config.get("split_by_batch"):
                    base, ext = os.path.splitext(os.path.basename(f))
                    batch_id = row["batch_id"]
                    new_fname = f"{base}_{batch_id}_{self.hex}{ext}"
                else:
                    new_fname = os.path.basename(f)

                analysis_results_path = os.path.join(
                    base_cod_path, "analysis_results", new_fname
                )
                analysis_results_paths.append(analysis_results_path)
            # in case of multiple files, join them with a comma
            row[path_key] = ", ".join(analysis_results_paths)
        else:
            # in case there is no file_path_name defined, use key as path_key
            row[path_key] = ", ".join(file_paths)

        return path_key

    def collect_info_from_lab_json(self) -> list[dict]:
        """Reads lab metadata from a JSON file and creates a list of dictionaries.
        Reads lab metadata from the specified JSON file and converts it into a list of dictionaries.
        This list is used to add the rest of the fields.

        Returns:
            json_lab_data: A list of dictionaries containing lab metadata (aka j_data).
        """
        method_name = f"{self.collect_info_from_lab_json.__name__}"
        try:
            json_lab_data = relecov_tools.utils.read_json_file(
                self.readlabmeta_json_file
            )
        except ValueError as e:
            error_msg = (
                f"Invalid lab-metadata json file: self.{self.readlabmeta_json_file}"
            )
            self.update_all_logs(
                method_name,
                "error",
                error_msg,
            )
            self.log_report.print_log_report(method_name, ["error"])
            raise ValueError(error_msg) from e
        return json_lab_data

    def get_sample_idx_col_pos(self, config_key: str) -> int:
        """Extract which column contain sample names for that specific file from config

        Args:
            config_key (str): Key of the configuration item in the software configuration JSON.

        Returns:
            sample_idx_col_pos: column number which contains sample names
        """
        try:
            sample_idx_col_pos = self.software_config[config_key]["sample_col_idx"] - 1
        except (KeyError, TypeError):
            sample_idx_col_pos = 0
            self.update_all_logs(
                "get_sample_idx_col_pos",
                "warning",
                f"No valid sample-index-column defined in '{self.software_name}.{config_key}'. Using default instead.",
            )
        return sample_idx_col_pos

    def extract_file(
        self,
        file: list[str],
        dest_folder: str,
        sample_name: str | None = None,
        path_key: str | None = None,
    ) -> bool:
        """Copy input file to the given destination, include sample name and key in log
        Args:
            file (list[str]): Paths of the files that are going to be copied
            dest_folder (str): Folder with files from batch of samples
            sample_name (str, optional): Name of the sample in metadata. Defaults to None.
            path_key (str, optional): Metadata field for the file. Defaults to None.
        Returns:
            bool: True if the process was successful, else False
        """
        dest_folder = os.path.join(dest_folder, "analysis_results")
        os.makedirs(dest_folder, exist_ok=True)
        for filepath in file:
            out_filepath = os.path.join(dest_folder, os.path.basename(filepath))
            if os.path.isfile(out_filepath):
                self.log.debug(f"{out_filepath} already exists, not extracted")
                continue
            if filepath == self.config_json.get_topic_data(
                "generic", "not_provided_field"
            ):
                self.update_all_logs(
                    self.extract_file.__name__,
                    "warning",
                    f"File for {path_key} not provided in sample {sample_name}",
                )
                continue
            try:
                shutil.copy(filepath, out_filepath)
            except (IOError, PermissionError) as e:
                self.update_all_logs(
                    self.extract_file.__name__,
                    "warning",
                    f"Could not extract {filepath}: {e}",
                )
                continue
        return True

    def split_data_by_batch(self, j_data: list[dict]) -> dict:
        """Split metadata from json for each batch of samples found according to folder location of the samples.
        Args:
            j_data (list(dict)): List of dictionaries, one per sample, including metadata for that sample
        Returns:
            data_by_batch (dict(list(dict))): Dictionary containing parts of j_data corresponding to each
            different folder with samples (batch) included in the original json metadata used as input
        """
        unique_batchs = {x.get("sequence_file_path_R1") for x in j_data}
        data_by_batch = {batch_dir: {} for batch_dir in unique_batchs}
        for batch_dir in data_by_batch:
            data_by_batch[batch_dir]["j_data"] = [
                samp
                for samp in j_data
                if samp.get("sequence_file_path_R1") == batch_dir
            ]
        return data_by_batch

    def split_tables_by_batch(
        self,
        files_found_dict: dict,
        file_tag: str,
        batch_data: list[dict],
        output_dir: str,
    ) -> None:
        """Filter table content to output a new table containing only the samples present in given metadata
        Args:
            files_found_dict (dict): A dictionary containing file paths identified for each configuration item.
            file_tag (str): File tag to be added to the new table file name.
            batch_data (list(dict)): Metadata corresponding to a single folder with samples (folder)
            output_dir (str): Output location for the generated tabular file
        """

        def extract_batch_rows_to_file(file, new_filename):
            """Create a new table file only with rows matching samples in batch_data"""
            ext_dict = {".csv": ",", ".tsv": "\t", ".tab": "\t"}
            file_extension = os.path.splitext(file)[1]
            file_df = pd.read_csv(
                file, sep=ext_dict.get(file_extension), header=header_pos
            )
            sample_col = file_df.columns[sample_col_pos]
            file_df[sample_col] = file_df[sample_col].astype(str)
            file_df = file_df[file_df[sample_col].isin(batch_samples)]

            os.makedirs(os.path.join(output_dir, "analysis_results"), exist_ok=True)
            output_path = os.path.join(output_dir, "analysis_results", new_filename)
            sep = ext_dict.get(file_extension, ",")
            file_df.to_csv(output_path, index=False, sep=sep)
            return

        method_name = self.split_tables_by_batch.__name__
        # sample names are created by concatenating sequencing_sample_id and unique_sample_id
        # if unique_sample_id is not present, it will use only sequencing_sample_id.
        batch_samples = [
            (
                f"{row.get('sequencing_sample_id')}_{row.get('unique_sample_id')}"
                if row.get("unique_sample_id")
                else row.get("sequencing_sample_id")
            )
            for row in batch_data
        ]
        for key, files in files_found_dict.items():
            if not self.software_config[key].get("split_by_batch"):
                continue
            header_pos = self.software_config[key].get("header_row_idx", 1) - 1
            sample_col_pos = self.get_sample_idx_col_pos(key)
            for file in files:
                try:
                    if self.software_config[key].get("filepath_name"):
                        filepath_key = self.software_config[key].get("filepath_name")
                        new_filename = os.path.basename(
                            list({row[filepath_key] for row in batch_data})[0]
                        )
                    else:
                        base, ext = os.path.splitext(os.path.basename(file))
                        new_filename = f"{base}_{file_tag}{ext}"
                    extract_batch_rows_to_file(file, new_filename)
                except Exception as e:
                    log_type = (
                        "error"
                        if self.software_config[key].get("required")
                        else "warning"
                    )
                    self.update_all_logs(
                        method_name,
                        log_type,
                        f"Could not create batch table for {file}: {e}",
                    )
        return

    def merge_metadata(self, batch_filepath: str, batch_data: list[dict]) -> list[dict]:
        """
        Merge metadata json if sample does not exist in the metadata file,
        or prompt the user to update if --update flag is provided and sample differs.

        Args:
            batch_filepath (str): Path to save the json file with the metadata.
            batch_data (list): A list of dictionaries containing metadata of the samples.

        Returns:
            merged_metadata (list): The updated list of metadata entries.
        """
        merged_metadata = relecov_tools.utils.read_json_file(batch_filepath)
        prev_metadata_dict = {
            item["sequencing_sample_id"]: item for item in merged_metadata
        }

        overwrite_all = False

        for item in batch_data:
            sample_id = item["sequencing_sample_id"]
            if sample_id in prev_metadata_dict:
                if prev_metadata_dict[sample_id] != item:
                    if self.update:
                        if not overwrite_all:
                            stderr.print(
                                f"[red]Sample '{sample_id}' has different metadata than the existing entry in {batch_filepath}."
                            )
                            response = Prompt.ask(
                                "[yellow]Do you want to overwrite this sample?",
                                choices=["y", "n", "all"],
                                default="n",
                            )
                            if response == "all":
                                overwrite_all = True
                                stderr.print(
                                    "[green]All subsequent samples will be overwritten automatically."
                                )
                                prev_metadata_dict[sample_id] = item
                            elif response == "y":
                                prev_metadata_dict[sample_id] = item
                                stderr.print(f"[green]Sample '{sample_id}' updated.")
                            else:
                                stderr.print(
                                    f"[blue]Skipping update for sample '{sample_id}'."
                                )
                        else:
                            prev_metadata_dict[sample_id] = item
                            stderr.print(f"[green]Sample '{sample_id}' updated (auto).")
                    else:
                        error_msg = f"Sample '{sample_id}' has different data in {batch_filepath} and new metadata. Can't merge."
                        stderr.print(f"[red]{error_msg}")
                        self.log.error(error_msg)
                        raise ValueError(error_msg)
            else:
                prev_metadata_dict[sample_id] = item

        merged_metadata = list(prev_metadata_dict.values())
        relecov_tools.utils.write_json_to_file(merged_metadata, batch_filepath)
        return merged_metadata

    def get_multiple_sample_files(self) -> list[str]:
        """
        Get a list of software configuration keys that have multiple samples.

        Returns:
            multiple_sample_files (list[str]): A list of keys from the software configuration
        """
        multiple_sample_files = []
        multiple_sample_files.extend(
            key
            for key in self.software_config.keys()
            if self.software_config[key].get("multiple_samples")
        )
        return multiple_sample_files

    def filter_properties(self, data: list[dict]) -> list[dict]:
        """
        Remove properties from bioinfo_metadata that are not in the schema properties.

        Args:
        data : list[dict]
            Dictionary with the bioinfo metadata

        Returns:
        data: list[dict]
            Bioinfo metadata json filtered
        """
        valid_keys = set(self.json_schema.get("properties", {}).keys())

        for sample in data:
            for k in list(sample.keys()):
                if k not in valid_keys:
                    sample.pop(k)
        return data

    def split_extra_json_data(
        self, extra_data: dict, batch_data: list[dict]
    ) -> tuple[list[dict], str]:
        """
        Split extra json data based on the sample names in the batch_data.

        Args:
        extra_data
            dict: List of dictionaries containing extra metadata that needs to be filtered based on the samples in
        batch_data
            list[dict]: List of dictionaries containing metadata for the current batch of samples.

        Returns:
        filtered_batch_data: list[dict]
            List of dictionaries containing extra metadata filtered based on the samples in batch_data.
        filename: str
            The file name of the first item in filtered_batch_data.
        """
        # Build the set of sample identifiers accepted for the extra data.
        # Accept both the combined sequencing_unique identifier and the unique id alone.
        sample_names_in_batch: set[str] = set()
        for sample_data in batch_data:
            seq_id = str(sample_data.get("sequencing_sample_id") or "").strip()
            unique_id = str(sample_data.get("unique_sample_id") or "").strip()
            if unique_id:
                sample_names_in_batch.add(unique_id)
                if seq_id:
                    sample_names_in_batch.add(f"{seq_id}_{unique_id}")
            elif seq_id:
                sample_names_in_batch.add(seq_id)

        # filter extra_data based on sample names in batch_data
        filtered_batch_data = [
            sample_data
            for sample_data in extra_data
            if sample_data["sample_name"] in sample_names_in_batch
        ]

        if not filtered_batch_data:
            self.update_all_logs(
                self.split_extra_json_data.__name__,
                "warning",
                "No extra metadata matched samples in current batch.",
            )
            filename = extra_data[0]["file_name"] if extra_data else "extra_data"
            return filtered_batch_data, filename

        # get the file name from the first item in filtered_batch_data
        filename = filtered_batch_data[0]["file_name"]

        return filtered_batch_data, filename

    def _validate_jdata_and_log_errors(self, out_path: str) -> bool:
        """Validate the j_data against the JSON schema and log errors if any.

        Args:
        out_path (str)
            Path where the validation results will be logged.

        Returns:
            bool: True if validation is successful, False otherwise.
        """
        self.j_data = self.filter_properties(self.j_data)

        valid_rows, invalid_rows = relecov_tools.validate.Validate.validate_instances(
            self.j_data, self.json_schema, "sequencing_sample_id"
        )

        for sample in valid_rows:
            self.logsum.feed_key(
                key=out_path, sample=sample.get("sequencing_sample_id")
            )

        if invalid_rows:
            unique_failed_samples = list(
                {
                    sample
                    for samples in invalid_rows["samples"].values()
                    for sample in samples
                }
            )
            for error_message, failed_samples in invalid_rows["samples"].items():
                field_with_error = invalid_rows["fields"][error_message]
                sample_list = "', '".join(failed_samples)
                error_text = f"{error_message} in field '{field_with_error}' for {len(failed_samples)} sample/s: '{sample_list}'"
                log_fn = (
                    self.logsum.add_error
                    if len(unique_failed_samples) == len(self.j_data)
                    else self.logsum.add_warning
                )
                log_fn(key=out_path, entry=error_text)
                for fail_samp in failed_samples:
                    self.logsum.add_error(
                        key=out_path, sample=fail_samp, entry=error_text
                    )

            if not self.soft_validation:
                self.parent_create_error_summary(
                    called_module="read-bioinfo-metadata", logs=self.logsum.logs
                )
                self.log.warning(
                    "Metadata was not completely validate, fix the errors or run with --soft_validation"
                )
                stderr.print(
                    "[red]Metadata was not completely validate, fix the errors or run with --soft_validation"
                )
                return False
        else:
            self.j_data = valid_rows
            self.log.info("Bioinfo json successfully validated.")
            stderr.print("[green] Bioinfo json successfully validated.")

        return True

    def _write_and_split_batches(
        self,
        files_found_dict: dict,
        batch_filepath: str,
        extra_json_data: list[dict],
    ) -> None:
        """Splits json data by batch and writes it to the output file.

        Args:
        files_found_dict
            A dictionary containing file paths identified for each configuration item.
        batch_filepath str
            Path to save the json file with the metadata.
        extra_json_data
            A list of dictionaries containing extra metadata that needs to be filtered based on the samples in
            batch_data.
        """
        self.j_data = self._write_or_merge_json(batch_filepath, self.j_data)
        self.log.info(f"Created complete batch json file: {batch_filepath}")
        stderr.print(f"Created complete batch json file: {batch_filepath}")

        data_by_batch = self.split_data_by_batch(self.j_data)
        for batch_dir, batch_dict in data_by_batch.items():
            batch_data = batch_dict["j_data"]
            if not batch_data:
                self.log.warning(
                    f"Data from batch {batch_dir} was completely empty. Skipped."
                )
                self.update_all_logs(
                    self.create_bioinfo_file.__name__,
                    "warning",
                    f"Data from batch {batch_dir} was completely empty. Skipped.",
                )
                continue

            lab_code = batch_data[0].get(
                "submitting_institution_id", batch_dir.split("/")[-2]
            )
            batch_date = batch_data[0].get("batch_id", batch_dir.split("/")[-1])
            file_tag = batch_date + "_" + self.hex

            self.log.info(f"Processing data from {batch_dir}")
            self.split_tables_by_batch(
                files_found_dict, file_tag, batch_data, batch_dir
            )

            batch_filename = self.tag_filename(
                "bioinfo_lab_metadata_" + lab_code + ".json"
            )
            batch_filepath = os.path.join(batch_dir, batch_filename)

            batch_data = self._write_or_merge_json(batch_filepath, batch_data)
            self.log.info(f"Created output laboratory json file: {batch_filepath}")
            stderr.print(f"Created output laboratory json file: {batch_filepath}")

            for sample in batch_data:
                self.logsum.feed_key(
                    key=batch_dir, sample=sample.get("sequencing_sample_id")
                )

            for extra_json in extra_json_data:
                filtered_batch_data, filename = self.split_extra_json_data(
                    extra_json, batch_data
                )
                extra_filename = f"{filename}_{lab_code}_{file_tag}.json"
                extra_filepath = os.path.join(batch_dir, extra_filename)
                relecov_tools.utils.write_json_to_file(
                    filtered_batch_data if filtered_batch_data else [], extra_filepath
                )

        self.parent_create_error_summary(
            called_module="read-bioinfo-metadata", logs=self.logsum.logs
        )

    def _write_or_merge_json(self, filepath: str, data: list[dict]) -> list[dict]:
        """
        Writes data to a JSON file, or merges with existing data if the file already exists.

        Args:
            filepath (str): Path to the output JSON file.
            data (list[dict]): List of dictionaries to write or merge.

        Returns:
            list[dict]: The resulting data after potential merge.
        """
        if os.path.exists(filepath):
            stderr.print(
                f"[blue]Bioinfo metadata {filepath} file already exists. Merging new data if possible."
            )
            self.log.info(
                f"Bioinfo metadata {filepath} file already exists. Merging new data if possible."
            )
            return self.merge_metadata(filepath, data)

        relecov_tools.utils.write_json_to_file(data, filepath)
        return data

    def create_bioinfo_file(self) -> bool:
        """Create the bioinfodata json with collecting information from lab
        metadata json, mapping_stats, and more information from the files
        inside input directory.

        Returns:
            bool: True if the bioinfo file creation process was successful.
        """

        tag = "bioinfo_lab_metadata_"
        file_tag = self.batch_id + "_" + self.hex
        out_filename = tag + file_tag + ".json"
        batch_filename = self.tag_filename(out_filename)

        year = str(datetime.now().year)
        out_path = os.path.join(self.output_dir, year)
        os.makedirs(out_path, exist_ok=True)
        batch_filepath = os.path.join(out_path, out_filename)

        self.validate_sample_names()

        stderr.print("[blue]Scanning input directory...")
        self.log.info("Scanning input directory")
        files_found_dict = self.scan_directory()

        stderr.print("[blue]Validating required files...")
        self.log.info("Validating required files")
        self.validate_software_mandatory_files(files_found_dict)

        stderr.print("[blue]Adding bioinfo metadata to read lab metadata...")
        self.log.info("Adding bioinfo metadata to read lab metadata")
        self.j_data, extra_json_data = self.add_bioinfo_results_metadata(
            files_found_dict, file_tag, self.j_data, out_path
        )

        stderr.print("[blue]Adding fixed values")
        self.log.info("Adding fixed values")
        self.j_data = self.add_fixed_values(self.j_data, batch_filename)

        stderr.print("[blue]Adding files path to read lab metadata")
        self.log.info("Adding files path to read lab metadata")
        self.j_data = self.map_and_extract_bioinfo_paths(files_found_dict, self.j_data)

        stderr.print("[blue]Evaluating quality control...")
        self.log.info("Evaluating quality control")
        module = importlib.import_module(
            f"relecov_tools.assets.pipeline_utils.{self.software_name}"
        )
        try:
            if hasattr(module, "quality_control_evaluation"):
                qc_func = getattr(module, "quality_control_evaluation")
                self.j_data = qc_func(self.j_data)
        except (AttributeError, NameError, TypeError, ValueError) as e:
            self.update_all_logs(
                self.create_bioinfo_file.__name__,
                "warning",
                f"Could not evaluate quality_control_evaluation for batch {self.j_data}: {e}",
            )
            stderr.print(
                f"[orange]Could not evaluate quality_control_evaluation for batch {self.j_data}: {e}"
            )

        stderr.print("[blue]Validating bioinfo metadata...")
        self.log.info("Validating bioinfo metadata")
        if not self._validate_jdata_and_log_errors(out_path):
            return False

        stderr.print("[blue]Writing and splitting batches...")
        self.log.info("Writing and splitting batches")
        self._write_and_split_batches(
            files_found_dict=files_found_dict,
            batch_filepath=batch_filepath,
            extra_json_data=extra_json_data,
        )

        return True



#!/usr/bin/env python

import sys
import re
import pandas as pd
import numpy as np
import json
import argparse
from geopy.geocoders import Nominatim
from geopy.exc import GeocoderTimedOut


def parse_args(args=None):
    Description = "Merge hospital databases with CCN and CODCNH into a unique database and a .json file"
    Epilog = """Example usage: python create_hospital_database.py -r DOCUMENTOS_ORIGINALES_2024/Metadatos_CARLOS_III_20250714.xlsx -c DOCUMENTOS_ORIGINALES_2024/CNH_2024_12122024.xlsx -o hospitals_ddbb.xlsx -a non_ddbb_adress.json -j regcess_laboratory_address.json -d differences.json -m missing.json -p laboratory_address.json"""
    parser = argparse.ArgumentParser(description=Description, epilog=Epilog)
    parser.add_argument(
        "-r",
        "--regcess",
        default="Metadatos_CARLOS_III_20250711.xlsx",
        help="Path to REGCESS file with the CCN code",
    )
    parser.add_argument(
        "-c",
        "--cnh",
        default="CNH_2024.xlsx",
        help="Path to ministry file of national hospital catalogue with the CODCNH code",
    )
    parser.add_argument(
        "-o",
        "--out_excel",
        default="hospitals_ddbb.xlsx",
        help="Output excel file path.",
    )
    parser.add_argument(
        "-p", "--previous_json", help="Previous .json file path", default=False
    )
    parser.add_argument(
        "-a",
        "--additional_json",
        default=False,
        help="Additional .json file with hospitals not included in the database.",
    )
    parser.add_argument(
        "-j",
        "--output_json",
        default="laboratory_address.json",
        help="Output .json filename",
    )
    parser.add_argument(
        "-d",
        "--differences_json",
        default="diferences.json",
        help="Json file with the keys that changed between the new and the previous dictionary",
    )
    parser.add_argument(
        "-m",
        "--missing_json",
        default="missing.json",
        help="Json file with the hospitals that were removed from the previous dictionary",
    )
    parser.add_argument(
        "-g",
        "--geo_loc_cities",
        default="geo_loc_cities.json",
        help="Json file with latitude and longitude of cities",
    )
    return parser.parse_args(args)


def process_regcess_table(file):
    """Load and clean the REGCESS Excel sheet into a standardized DataFrame.

    Steps:
        1. Read "Metadatos" sheet as DataFrame (CCN, Cód. Centro Autonómico, etc. as str).
        2. Ensure "Cód. Clase de Centro" column exists (fill NaN if missing).
        3. Transform cases: Title Case: Nombre Centro & Dependencia Funcional; lowercase: Email.
        4. Build "Dirección" by concatenating Tipo Vía, Nombre Vía, Número Vía.
        5. Split "Coordenadas" into Latitud and Longitud columns.
        6. Clean commas and whitespace in text columns.
        7. Normalize inverted names ("Coruña, A" → "A Coruña").
        8. Return cleaned DataFrame.

    Args:
        file_path (str): path to the REGCESS Excel file

    Returns:
        regcess_db (pd.DataFrame): cleaned REGCESS table
    """

    # 1. Load with proper dtypes
    regcess_db = pd.read_excel(
        file,
        sheet_name="Metadatos",
        dtype={
            "CCN": str,
            "Cód. Centro Autonómico": str,
            "Código Postal": str,
            "Teléfono Principal": str,
        },
    )

    # 2. Ensure missing column
    if "Cód. Clase de Centro" not in regcess_db.columns:
        regcess_db["Cód. Clase de Centro"] = np.nan

    # 3. Transform case
    regcess_db["Nombre Centro"] = regcess_db["Nombre Centro"].str.title()
    regcess_db["Dependencia Funcional"] = regcess_db[
        "Dependencia Funcional"
    ].str.title()
    regcess_db["Email"] = regcess_db["Email"].str.lower()
    regcess_db["Email"] = regcess_db["Email"].fillna("Desconocido")

    # 4. Construct Dirección
    regcess_db["Dirección"] = (
        regcess_db["Tipo Vía"].str.title().str.strip()
        + " "
        + regcess_db["Nombre Vía"].str.title().str.strip()
        + ", "
        + regcess_db["Número Vía"].astype(str).str.strip()
    )

    # 5. Split Latitude and Longitude in two different columns
    regcess_db["Coordenadas"] = regcess_db["Coordenadas"].astype(str)
    regcess_db[["Latitud", "Longitud"]] = regcess_db["Coordenadas"].str.split(
        ",", expand=True
    )

    # 6. Clean text columns (remove double witespaces, double commas and double S/Ns)
    regcess_db["Nombre Centro"] = regcess_db["Nombre Centro"].apply(
        lambda x: re.sub(r"\s+", " ", x) if isinstance(x, str) else x
    )
    regcess_db["Dependencia Funcional"] = regcess_db["Dependencia Funcional"].apply(
        lambda x: re.sub(r"\s+", " ", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r",,+", ",", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r",\s*,", ",", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r"\s+", " ", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r",\s*S/N(,\s*S/N)+", ", S/N", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r"\s*S/N(,\s*S/N)+", ", S/N", x) if isinstance(x, str) else x
    )
    regcess_db["Dirección"] = regcess_db["Dirección"].apply(
        lambda x: re.sub(r"\s*Sn(,\s*S/N)+", ", S/N", x) if isinstance(x, str) else x
    )

    # 7. Normalize names
    regcess_db["Municipio"] = regcess_db["Municipio"].apply(normalize_names)
    regcess_db["Municipio"] = regcess_db["Municipio"].str.title()
    regcess_db["CCAA"] = regcess_db["CCAA"].apply(normalize_names)
    regcess_db["Provincia"] = regcess_db["Provincia"].apply(normalize_names)

    # 8. Fix some hospital names
    regcess_db["Nombre Centro"] = regcess_db["Nombre Centro"].apply(
        lambda x: center_name_map(x) if center_name_map(x) is not None else x
    )

    return regcess_db


def process_cnh_table(file):
    """Load and clean the CNH hospital catalogue sheet.
    Args:
        file_path (str): path to CNH Excel file
    Returns:
        cnh_ddbb (pd.DataFrame): cleaned CNH DataFrame
    """
    cnh_ddbb = pd.read_excel(
        file,
        sheet_name="DIRECTORIO DE HOSPITALES",
        dtype={
            "CCN": str,
            "CODCNH": str,
            "CODIDCOM": str,
            "Cód. Municipio": str,
            "Cód. Provincia": str,
            "Cód. CCAA": str,
        },
    )
    cnh_ddbb["Nombre del Complejo"] = cnh_ddbb["Nombre del Complejo"].str.title()

    return cnh_ddbb


def process_complex(cnh_ddbb, regcess_db):
    """Extract complex hospitals, merge data, and standardize for JSON creation.

    Steps:
        1. Filter CNH entries with 'Forma parte Complejo' == 'S'.
        2. Select relevant columns for complexes.
        3. Merge with regcess_db on 'CCN', inner join.
        4. Drop duplicates by 'CODIDCOM', keep first.
        5. Prefix new CCN with 'COMP_' + CODIDCOM.
        6. Overwrite/hide fields not applicable to complexes:
           - Set 'Cód. Centro Autonómico', 'Dependencia Funcional', 'Cód. Clase de Centro' to 'N/A'
           - Set 'Clase de Centro' to 'Complejo'
        7. Drop 'CODIDCOM' column, rename 'Nombre del Complejo' to 'Nombre Centro'.
        8. Return the processed DataFrame.

    Args:
        cnh_ddbb (pd.DataFrame): CNH DataFrame
        regcess_db (pd.DataFrame): REGCESS DataFrame

    Returns:
        complex_hospitals (pd.DataFrame): DataFrame of complex hospitals
    """
    complex_hospitals = cnh_ddbb[cnh_ddbb["Forma parte Complejo"] == "S"]
    complex_hospitals = complex_hospitals[
        [
            "CCN",
            "CODCNH",
            "Nombre del Complejo",
            "CODIDCOM",
            "Cód. Municipio",
            "Cód. Provincia",
            "Cód. CCAA",
        ]
    ]
    complex_hospitals = pd.merge(
        complex_hospitals,
        regcess_db,
        on=["CCN"],
        how="inner",
        suffixes=("_compl", "_regcess"),
    )

    # Remove hospital name from REGCESS
    complex_hospitals.drop(columns=["Nombre Centro"], inplace=True)
    complex_hospitals["CODCNH"] = complex_hospitals["CODIDCOM"]

    # keep only last hospital for information (ESRI one of the hospitals, randomly)
    complex_hospitals = complex_hospitals.drop_duplicates(
        subset="CODIDCOM", keep="first"
    )

    # Remove hospital info not aplicable to complex
    # Replace CCN which is assoaciated with hospital and not the complex, with a combination of COMP_ + the complex code CODIDCOM
    complex_hospitals["CCN"] = "COMP_" + complex_hospitals["CODIDCOM"]
    # Remove data from the hospital
    complex_hospitals[
        ["Cód. Centro Autonómico", "Dependencia Funcional", "Cód. Clase de Centro"]
    ] = "N/A"
    # Replaced "Clase de Centro" with "Complejo" instead of the hospital's values.
    complex_hospitals[["Clase de Centro"]] = "Complejo"

    # Add "Complejo" to the "Nombre del Complejo" if they don't have it
    sin_complejo = ~complex_hospitals["Nombre del Complejo"].str.contains(
        "Complejo", na=False
    )
    complex_hospitals.loc[sin_complejo, "Nombre del Complejo"] = (
        "Complejo " + complex_hospitals.loc[sin_complejo, "Nombre del Complejo"]
    )

    # Fix column names and numbers to fit in the hospital table names
    complex_hospitals = complex_hospitals.drop(columns=["CODIDCOM"]).rename(
        columns={"Nombre del Complejo": "Nombre Centro"}
    )

    return complex_hospitals


def normalize_names(name):
    """Swap inverted names around a comma and normalize slash-separated parts.
    Args:
        name (str): raw location string, e.g. "Coruña, A" or "Vila Joiosa, la/Villajoyosa"
    Returns:
        str: normalized name, e.g. "A Coruña" or "la Vila Joiosa/Villajoyosa"
    """

    if "," in name:
        if "/" not in name:
            first, second = [p.strip() for p in name.split(",", 1)]
            return f"{second} {first}"
        else:
            parts = name.split("/")
            normalized = []
            for part in parts:
                part = part.strip()
                if "," in part:
                    first, second = [p.strip() for p in part.split(",", 1)]
                    normalized.append(f"{second} {first}")
                else:
                    normalized.append(part)
        return "/".join(normalized)
    else:
        return name


def create_json(hospitals):
    """Convert hospital DataFrame to a JSON dict.
    Args:
        hospitals_df (pd.DataFrame): DataFrame with hospital records

    Returns:
        hospitals_json (dict): hospital info dict
    """
    hospitals_json = {}

    for _, row in hospitals.iterrows():

        if not pd.isna(row["Cód. Clase de Centro"]):
            center_class_code = row["Cód. Clase de Centro"].strip()
        else:
            class_name = row["Clase de Centro"].strip()
            center_class_code = center_class_map(class_name)

        ccn_hospital = row["CCN"]
        hospitals_json[ccn_hospital] = {
            "collecting_institution_code_1": ccn_hospital,
            "collecting_institution_code_2": row["CODCNH"],
            "collecting_institution": row["Nombre Centro"].strip(),
            "collecting_institution_address": row["Dirección"],
            "collecting_institution_email": (
                row["Email"].strip() if not pd.isna(row["Email"]) else "Desconocido"
            ),
            "autonom_cod": row["Cód. Centro Autonómico"],
            "geo_loc_state": row["CCAA"].strip(),
            "geo_loc_state_cod": row["Cód. CCAA"],
            "geo_loc_region": row["Provincia"].strip(),
            "geo_loc_region_cod": row["Cód. Provincia"],
            "geo_loc_city": row["Municipio"].strip(),
            "geo_loc_city_cod": row["Cód. Municipio"],
            "post_code": row["Código Postal"],
            "dep_func": row["Dependencia Funcional"].strip(),
            "center_class_code": center_class_code,
            "collecting_institution_function": row["Clase de Centro"].strip(),
            "lab_geo_loc_latitude": row["Latitud"],
            "lab_geo_loc_longitude": row["Longitud"],
            "collecting_institution_phone": row["Teléfono Principal"],
            "geo_loc_country": "Spain",
        }

    return hospitals_json


def center_name_map(center_name):
    """Map center names with wrong accents ino the propper name.
    Args:
        center_name (str): e.g. "GERMÃ?N SÃ?NCHEZ LÃ?PEZ"
    Returns:
        center_name_fixed (str): "GERMÁN SÁNCHEZ LÓPEZ"
    """
    center_map = {
        "Hospital FãTima": "Hospital Fatima",
        "Hospital Mare De Dã?A©U De La Mercã?A?": "Hospital Mare De Deu De La Merce",
        "Red De Salud Mental De Araba (Hospital PsiquiãTrico De Araba)": "Red De Salud Mental De Araba (Hospital Psiquiatrico De Araba)",
    }
    center_name_fixed = center_map.get(center_name)
    return center_name_fixed


def center_class_map(class_name):
    """Map human-readable center class name to its code.
    Args:
        class_name (str): e.g. "Centros de Diagnostico"
    Returns:
        center_class_code (str): code like "C256"
    """
    class_map = {
        "Hospitales Generales": "C11",
        "Hospitales especializados": "C12",
        "Hospitales de media y larga estancia": "C13",
        "Hospitales de salud mental y tratamiento de toxicomanías": "C14",
        "Otros Centros con Internamiento": "C190",
        "Consultas Médicas": "C21",
        "Centros de salud": "C231",
        "Consultorios de atencion primaria": "C232",
        "Centros de Diagnostico": "C256",
        "Centros moviles de asistencia sanitaria": "C257",
    }
    center_class_code = class_map.get(class_name)
    return center_class_code


def add_hospitals(hospital_ddbb_json, regcess_db, add_json):
    """Merge additional hospital entries into existing JSON, tracking differences, and updates additional hospitals with REGCESS information.
    Args:
        hospital_json (dict): existing hospital dict
        regcess_db (pd.DataFrame): REGCESS DataFrame
        add_json (dict): new hospital entries
    Returns:
        tuple: (updated_hospital_json, updated_add_json, differences)
          - hospital_ddbb_json: JSON dict with the merged list of hospitals
          - add_json: JSON dict wih the additional hospitals updated with REGCESS
          - differences: JSON dict with the differences found between add_json and regcess_db
    """

    differences = {}

    for ccn, data in add_json.items():
        if ccn in regcess_db["CCN"].values:
            row = regcess_db[regcess_db["CCN"] == ccn].iloc[0]
            if not pd.isna(row["Cód. Clase de Centro"]):
                center_class_code = row["Cód. Clase de Centro"].strip()
            else:
                class_name = row["Clase de Centro"].strip()
                center_class_code = center_class_map(class_name)

            if row["CCAA"].strip() == data.get("geo_loc_state"):
                geo_loc_state_cod = data.get("geo_loc_state_cod")
            else:
                geo_loc_state_cod = "ADD"
            if row["Provincia"].strip() == data.get("geo_loc_region"):
                geo_loc_region_cod = data.get("geo_loc_region_cod")
            else:
                geo_loc_region_cod = "ADD"
            if row["Municipio"].strip() == data.get("geo_loc_city"):
                geo_loc_city_cod = data.get("geo_loc_city_cod")
            else:
                geo_loc_city_cod = "ADD"

            new_data = {
                "collecting_institution_code_1": data.get(
                    "collecting_institution_code_1"
                ),
                "collecting_institution_code_2": data.get(
                    "collecting_institution_code_2"
                ),
                "collecting_institution": row["Nombre Centro"].strip(),
                "collecting_institution_address": row["Dirección"],
                "collecting_institution_email": (
                    row["Email"].strip()
                    if not pd.isna(row.get("Email"))
                    else "Desconocido"
                ),
                "autonom_cod": row["Cód. Centro Autonómico"],
                "geo_loc_state": row["CCAA"].strip(),
                "geo_loc_state_cod": geo_loc_state_cod,
                "geo_loc_region": row["Provincia"].strip(),
                "geo_loc_region_cod": geo_loc_region_cod,
                "geo_loc_city": row["Municipio"].strip(),
                "geo_loc_city_cod": geo_loc_city_cod,
                "post_code": row["Código Postal"],
                "dep_func": row["Dependencia Funcional"].strip(),
                "center_class_code": center_class_code,
                "collecting_institution_function": row["Clase de Centro"].strip(),
                "lab_geo_loc_latitude": row["Latitud"],
                "lab_geo_loc_longitude": row["Longitud"],
                "collecting_institution_phone": row["Teléfono Principal"],
                "geo_loc_country": "Spain",
            }

            for field, new_val in new_data.items():
                old_val = (
                    data.get(field, "").strip()
                    if isinstance(data.get(field), str)
                    else data.get(field)
                )
                if old_val != new_val:
                    if field not in differences:
                        differences[field] = {}
                    differences[field].update({old_val: new_val})

            add_json[ccn] = new_data
            hospital_ddbb_json[ccn] = new_data

        else:
            hospital_ddbb_json[ccn] = data

    return hospital_ddbb_json, add_json, differences


def compare_json(prev_json_path, new_hospitals, differences):
    """Compare previous JSON with new entries, find missing and changed fields.
    Args:
        prev_json_path (str): path to previous JSON file
        new_hospitals (dict): updated hospital dict
        differences (dict): existing differences tracker
    Returns:
        tuple: (missing_json, differences)
          - missing_json: JSON dict of hospitals that were removed
          - differences: JSON dict with the differences found between the old and the new hospital list
    """
    missing_json = {}

    with open(prev_json_path, "r", encoding="utf-8") as json_file:
        previous_json = json.load(json_file)

    for prev_ccn, prev_data in previous_json.items():
        data_found = False
        for ccn, data in new_hospitals.items():
            if ccn == prev_ccn:
                data_found = True
                for field in data:
                    if field in prev_data:
                        if data[field] != prev_data[field]:
                            if field not in differences:
                                differences[field] = {}
                            differences[field].update({prev_data[field]: data[field]})
                break
        if not data_found:
            missing_json[prev_ccn] = prev_data

    return missing_json, differences


def write_json(output_json, data):
    """Write a dict to a JSON file with pretty formatting and message.
    Args:
        output_path (str): filename for JSON output
        data (dict): dictionary to print
    """
    with open(output_json, "w", encoding="utf-8") as json_file:
        json.dump(data, json_file, ensure_ascii=False, indent=4)

    print(f"Archivo JSON generado: {output_json}")


def create_cities_coord(hospital_json, geo_loc_file):
    """Ensure all cities have coordinates and build summary strings.
    Args:
        hospital_json (dict): hospital info dict
        geo_loc_file (str): path to JSON with existing coordinates
    """
    with open(geo_loc_file, "r", encoding="utf-8") as json_file:
        geo_loc_json = json.load(json_file)

    geo_loc_cities = set(geo_loc_json.keys())

    cities = []
    states = []
    regions = []
    hospitals = []
    geolocator = Nominatim(user_agent="my_app")
    for _, data in hospital_json.items():
        loc = None
        city = data["geo_loc_city"]
        if city not in cities:
            cities.append(city)
        state = data["geo_loc_state"]
        if state not in states:
            states.append(state)
        region = data["geo_loc_region"]
        if region not in regions:
            regions.append(region)
        hospital = data["collecting_institution"]
        if hospital not in hospitals:
            hospitals.append(hospital)

        country = data["geo_loc_country"]
        if city not in geo_loc_cities:
            try:
                loc = geolocator.geocode(f"{city}, {state}, {country}", timeout=15)
            except GeocoderTimedOut:
                return None, None
            if loc:
                geo_loc_json[data["geo_loc_city"]] = {
                    "geo_loc_latitude": str(loc.latitude),
                    "geo_loc_longitude": str(loc.longitude),
                }
            else:
                print("City with coordenates not found")
                print(f"{city}, {state}, {country}")

    write_json(geo_loc_file, geo_loc_json)

    schema_data_json = {}
    schema_data_json["cities"] = "; ".join(cities)
    schema_data_json["states"] = "; ".join(states)
    schema_data_json["regions"] = "; ".join(regions)
    schema_data_json["hospitals"] = "; ".join(hospitals)

    write_json("schema_excel_strings.json", schema_data_json)

    return


def main(args=None):
    args = parse_args(args)

    # Load regcess table and transform data
    regcess_db = process_regcess_table(args.regcess)

    # Load ministry table and transform/merge data
    cnh_ddbb = process_cnh_table(args.cnh)

    # Remove columns that we are not using
    regcess_db.drop(
        columns=["Unnamed: 10", "Coordenadas", "Tipo Vía", "Nombre Vía", "Número Vía"],
        errors="ignore",
        inplace=True,
    )

    cod_mapping = cnh_ddbb[
        ["CCN", "CODCNH", "Cód. Municipio", "Cód. Provincia", "Cód. CCAA"]
    ]

    # Merge both tables
    hospital_ddbb = pd.merge(
        regcess_db, cod_mapping, on=["CCN"], how="inner", suffixes=("_regcess", "_cnh")
    )

    # Process hospital complex
    complex_table = process_complex(cnh_ddbb, regcess_db)

    hospital_ddbb = pd.concat([hospital_ddbb, complex_table], ignore_index=True)

    hospital_ddbb_json = create_json(hospital_ddbb)

    if args.additional_json:
        with open(args.additional_json, "r", encoding="utf-8") as json_file:
            add_json = json.load(json_file)
        hospital_ddbb_json, add_json, differences = add_hospitals(
            hospital_ddbb_json, regcess_db, add_json
        )

    write_json(args.additional_json, add_json)

    hospital_ddbb = pd.DataFrame.from_dict(hospital_ddbb_json, orient="index")
    hospital_ddbb.reset_index(inplace=True)
    hospital_ddbb.rename(columns={"index": "ccn"}, inplace=True)

    hospital_ddbb.to_excel(args.out_excel, sheet_name="hospitals", index=False)
    print(f"Archivo Excel generado: {args.out_excel}")

    if args.previous_json:
        missing_json, differences = compare_json(
            args.previous_json, hospital_ddbb_json, differences
        )
        if len(missing_json) != 0:
            write_json("missing_hospitals.json", missing_json)

    write_json(args.output_json, hospital_ddbb_json)

    if len(differences) != 0:
        write_json(args.differences_json, differences)

    create_cities_coord(hospital_ddbb_json, args.geo_loc_cities)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
import logging
import requests
import rich.console
from typing import Any
import relecov_tools.utils
from requests.auth import HTTPBasicAuth

log = logging.getLogger(__name__)

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class RestApi:
    def __init__(self, server, url):
        self.request_url = server + url
        self.headers = {"content-type": "application/json"}
        self.UNABLE_TO_CONNECT = {
            "ERROR": "Server not available",
            "status_code": "503 Service Unavailable",
            "data": {},
        }

    def get_request(
        self,
        request_info,
        parameter=None,
        value=None,
        safe=True,
        credentials=None,
        params={},
    ):
        """Send a GET request with optional parameters and authentication.

        Args:
            request_info (str): Endpoint path appended to the base URL.
            parameter ([str, dict, None]): Query parameter(s). If a string, used with `value`.
            If given as a dict, key-value pairs are joined as query string. If empty or None, no parameters.
            value (Optional[str]): Value for the `parameter` if it is a string.
            safe (bool): If True, errors are logged and not raised. Defaults to True.
            credentials ([dict, tuple, list, None]): Basic auth credentials. Can be a dict with
            'user' and 'pass' as keys, or a tuple/list (user, pass).
            params (dict): dictionary of params for the get request

        Returns:
            dict: Same as RestApi.standardize_response() dict
        """
        if parameter == "" or parameter is None:
            url_http = str(self.request_url + request_info)
        elif isinstance(parameter, dict):
            param_value = []
            for key, value in parameter.items():
                param_value.append(key + "=" + value)
            url_http = str(
                self.request_url + request_info + "?" + "&".join(param_value)
            )
        else:
            url_http = str(
                self.request_url + request_info + "?" + parameter + "=" + value
            )
        if credentials is not None:
            if isinstance(credentials, dict):
                credentials = HTTPBasicAuth(credentials["user"], credentials["pass"])
            elif isinstance(credentials, (list, tuple)):
                credentials = HTTPBasicAuth(credentials[0], credentials[1])
        try:
            response = requests.get(
                url_http, headers=self.headers, auth=credentials, params=params
            )
            return RestApi.standardize_response(
                response, success_codes=[200, 201], method="GET"
            )
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return self.UNABLE_TO_CONNECT

    def put_request(self, data, credentials, url):
        """Send a PUT request to update data on the server.

        Args:
            data (dict): Data payload for the PUT request.
            credentials (dict): Basic auth credentials, with 'user' and 'pass' keys.
            url (str): Endpoint path appended to the base URL.

        Returns:
            dict: Same as RestApi.standardize_response() dict
        """
        if isinstance(credentials, dict):
            auth = (credentials["user"], credentials["pass"])
        url_http = str(self.request_url + url)
        try:
            response = requests.put(url_http, data=data, auth=auth)
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return self.UNABLE_TO_CONNECT
        return RestApi.standardize_response(response, success_codes=[201], method="PUT")

    def post_request(self, data, credentials, url, file=None):
        """Send a POST request with optional file upload.

        Args:
            data (dict): Data payload for the POST request.
            credentials (dict): Basic auth credentials, with 'user' and 'pass' keys.
            url (str): Endpoint path appended to the base URL.
            file (Optional[str]): Path to a file to be uploaded (if any).

        Returns:
            dict: Same as RestApi.standardize_response() dict
        """
        if isinstance(credentials, dict):
            auth = (credentials["user"], credentials["pass"])
        url_http = str(self.request_url + url)
        try:
            if file:
                files = {"upload_file": open(file, "rb")}
                response = requests.post(
                    url_http, files=files, data=data, headers=self.headers, auth=auth
                )
            else:
                response = requests.post(
                    url_http, data=data, headers=self.headers, auth=auth
                )
            return RestApi.standardize_response(
                response, success_codes=[201], method="POST"
            )
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return self.UNABLE_TO_CONNECT

    def sample_already_in_db(self, api_func, credentials, sample_data):
        """Check if sample with data already exists in the target platform

        Args:
            api_func (str): Api functionality to check if sample is present
            credentials (dict["user": user, "pass": pass]): Credentials dictionary
            sample_data (dict): dictionary with sample metadata to use as input

        Raises:
            ValueError: if the conection was interrupted unexpectedly.

        Returns:
            bool: Wether the sample is found in the database or not.
        """
        response = self.get_request(
            api_func, credentials=credentials, params=sample_data, safe=False
        )
        log.info(str(response))
        if missing_keys := [
            key for key in ["message", "data", "status_code"] if key not in response
        ]:
            log.error(f"Missing keys in API response: {missing_keys}")
            stderr.print(f"Missing keys in API response: {missing_keys}")
            raise KeyError(f"Missing keys in API response: {missing_keys}")

        if response["status_code"] == 404:
            return False
        elif response["status_code"] == 200:
            return not bool(
                response["message"] != "Sample not found" or response["data"]
            )
        else:
            raise ValueError(f"Error trying to check for sample: {response}")

    @staticmethod
    def standardize_response(
        response,
        success_codes: list[int] = [200],
        method: str = "",
        *,
        message_keys: list[str] = ["message", "detail", "msg"],
        error_keys: list[str] = ["ERROR", "error"],
        data_keys: list[str] = ["data", "payload", "result"],
        include_raw_on_non_json: bool = False,
    ) -> dict[str, Any]:
        """
        Normalize a `requests.Response` into your canonical shape while preserving API-provided fields.

        Canonical output:
            {
              "message": <str>,                # Prefer API 'message'/'detail'/...
              "status_code": <int>,            # Always present
              "data": <obj>,                   # Prefer API 'data'/'payload'/'result', else {}
              "ERROR": <str> (optional)        # Included if API provided it, or if status is not success
              ["raw"]: <str> (optional)        # Raw text for non-JSON responses if enabled
            }

        Flexibility:
          - Accepts alternative key names via message_keys, error_keys, data_keys.
          - If the API already provides 'message', 'data', 'ERROR', those are preserved.
          - If response is not JSON, still returns the canonical keys.
        """
        status_code = getattr(response, "status_code", None)
        content_type = (
            response.headers.get("Content-Type", "")
            if hasattr(response, "headers")
            else ""
        )
        is_success = status_code in success_codes

        # Try to parse JSON safely
        payload: Any = None
        if "application/json" in content_type.lower():
            try:
                payload = response.json()
            except Exception:
                payload = None
        else:
            # Some APIs forget the content-type; try JSON parse as best effort
            try:
                payload = response.json()
            except Exception:
                payload = None

        # Initialize canonical result
        result: dict[str, Any] = {
            "message": "",
            "status_code": status_code,
            "data": {},
        }

        def _first_key(d: dict[str, Any], keys: list[str]):
            """Get the first key from `keys` that is present in `d`

            Args:
            d : dict
                d is a dict to search
            keys
                list of keys to search

            Returns:
            key, value : tuple  returns the first key and value
            """
            for k in keys:
                if isinstance(d, dict) and k in d and d[k] not in (None, ""):
                    return k, d[k]
            return None, None

        if isinstance(payload, dict):
            # Prefer exact API fields if present
            # ERROR
            err_key, err_val = _first_key(payload, list(error_keys))
            if err_key is not None:
                result["ERROR"] = err_val

            # message - prefer API 'message'/'detail'/...
            if "message" in payload and payload["message"] is not None:
                result["message"] = payload["message"]
            else:
                _, alt_msg = _first_key(payload, list(message_keys))
                if alt_msg is not None:
                    result["message"] = alt_msg
                elif "ERROR" in result:
                    result["message"] = str(result["ERROR"])  # fallback to error text
                else:
                    result["message"] = "OK" if is_success else "Unexpected error"

            # data - prefer API 'data'/'payload'/'result'
            if "data" in payload:
                result["data"] = payload["data"]
            else:
                _, alt_data = _first_key(payload, list(data_keys))
                if alt_data is not None:
                    result["data"] = alt_data
                else:
                    # fallback: use all non-canonical keys
                    canon_keys = set(message_keys) | set(error_keys) | set(data_keys)
                    canon_keys |= {"message", "ERROR", "data"}
                    extra = {k: v for k, v in payload.items() if k not in canon_keys}
                    if extra:
                        result["data"] = extra
                    else:
                        result["data"] = {}

            # If HTTP status is not success and API did not include ERROR, synthesize it
            if not is_success and "ERROR" not in result:
                result["ERROR"] = result["message"] or f"HTTP {status_code}"

        elif isinstance(payload, list):
            # JSON list payload (e.g., sample-info returning rows). Keep as data.
            if is_success:
                result["message"] = (
                    "OK"
                    if not isinstance(payload, dict)
                    else result.get("message", "OK")
                )
                result["data"] = payload
            else:
                # On error with list (unlikely), still surface as data and set message
                result["message"] = "Unexpected error"
                result["data"] = payload

        else:
            # Non-JSON response
            text = ""
            try:
                text = response.text or ""
            except Exception:
                text = ""

            if not is_success:
                # On error, surface the body as ERROR (truncated to avoid huge logs)
                truncated = text.strip()[:1000]
                result["ERROR"] = truncated or "Unexpected error"
                result["message"] = truncated or "Unexpected error"
            else:
                # On success without JSON, be conservative
                result["message"] = "OK"
                result["data"] = {}

            if include_raw_on_non_json:
                result["raw"] = text

        # Logging
        if not is_success:
            logtxt = (
                f"Unable to {method} parameters. Received '{result.get('ERROR','Unexpected error')}' "
                f"with status code: {status_code}"
            )
            log.error(logtxt)
        else:
            logtxt = (
                f"Successful {method} response. Received '{result.get('message','OK')}' "
                f"with status code: {status_code}"
            )
            log.info(logtxt)

        return result



#!/usr/bin/env python3
from __future__ import annotations
import argparse
import csv
import datetime
import json
import logging
import os
import re
import unicodedata
from decimal import Decimal, ROUND_HALF_UP
from pathlib import Path
from typing import Any, Dict, Set

import mysql.connector

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

# ────── Constantes de BD ───────────────────────────────────────────────────
CITY_TABLE, CITY_COL_ID, CITY_COL_NAME = "core_city", "id", "city_name"
LAB_TABLE = "core_lab_request"
LAB_COL_ID, LAB_COL_CODE1, LAB_COL_CODING = "id", "lab_code_1", "lab_name_coding"

FIELD_MAP = {
    "collecting_institution_code_1": "lab_code_1",
    "collecting_institution_code_2": "lab_code_2",
    "collecting_institution": "lab_name",
    "collecting_institution_address": "address",
    "collecting_institution_email": "lab_email",
    "collecting_institution_phone": "lab_phone",
    "lab_geo_loc_latitude": "lab_geo_loc_latitude",
    "lab_geo_loc_longitude": "lab_geo_loc_longitude",
    "autonom_cod": "autonom_cod",
    "post_code": "post_code",
    "dep_func": "dep_func",
    "center_class_code": "center_class_code",
    "collecting_institution_function": "lab_function",
}
DEFAULTS = {"lab_unit": "-", "lab_contact_name": "-"}
DEFAULT_APPS = "wetlab"
MAX_CODING_LEN = 50


def _quote(column: str) -> str:
    """Wrap a column name in back-ticks for MySQL."""
    return f"`{column}`"


# ────── Generic helpers ──────────────────────────────────────────────────
def load_json(p: str | Path) -> Any:
    return json.loads(Path(p).read_text(encoding="utf-8"))


def rounded(v, nd=6):
    if v is None or str(v).strip().lower() in {"", "nan", "null", "none"}:
        return None
    try:
        return float(Decimal(str(v)).quantize(Decimal(f"1e-{nd}"), ROUND_HALF_UP))
    except Exception:
        logging.warning("Lat/Lon '%s' no numérico; NULL", v)
        return None


def normalize(txt: str) -> str:
    txt = unicodedata.normalize("NFKD", txt).encode("ascii", "ignore").decode()
    return re.sub(r"\s+", " ", txt).strip().lower()


# ────── unique lab_name_coding ─────────────────────────────────────────────
def clean_words(n: str) -> list[str]:
    n = re.sub(r"\([^)]*\)", " ", n)
    n = unicodedata.normalize("NFKD", n).encode("ascii", "ignore").decode()
    return [w for w in re.split(r"[^A-Za-z0-9]+", n) if w]


def build_code(words: list[str], k: int) -> str:
    return (
        "".join(
            w[0].upper() + w[1:k].lower() if len(w) >= k else w[0].upper()
            for w in words
        )
        or "LAB"
    )


def unique_code(name: str, taken: Set[str]) -> str:
    ws = clean_words(name)
    longest = max((len(w) for w in ws), default=1)
    for k in range(1, longest + 1):
        c = build_code(ws, k)[:MAX_CODING_LEN]
        if c not in taken:
            taken.add(c)
            return c
    base = build_code(ws, 1)[: MAX_CODING_LEN - 4]
    for n in range(1, 1000):
        c = f"{base}{n:03d}"[:MAX_CODING_LEN]
        if c not in taken:
            taken.add(c)
            return c
    raise RuntimeError("Unable to create unique lab_name_coding")


# ────── caches FK y codings ───────────────────────────────────────────────
def cache_cities(cur) -> Dict[str, int]:
    cur.execute(
        f"SELECT {_quote(CITY_COL_NAME)},{_quote(CITY_COL_ID)} FROM {_quote(CITY_TABLE)}"
    )
    return {r[CITY_COL_NAME]: r[CITY_COL_ID] for r in cur.fetchall()}


def cache_codings(cur) -> Set[str]:
    cur.execute(f"SELECT {_quote(LAB_COL_CODING)} FROM {_quote(LAB_TABLE)}")
    return {r[LAB_COL_CODING] for r in cur.fetchall() if r[LAB_COL_CODING]}


# ────── MAIN ───────────────────────────────────────────────────────────────
def main(a):
    data = load_json(a.lab_json)
    conn = mysql.connector.connect(
        host=a.host,
        port=a.port,
        user=a.user,
        password=a.password,
        database=a.database,
        autocommit=False,
        use_pure=True,
    )
    cur = conn.cursor(dictionary=True, buffered=True)
    cities = cache_cities(cur)
    taken = cache_codings(cur)

    q_lab, q_id, q_code1 = map(_quote, (LAB_TABLE, LAB_COL_ID, LAB_COL_CODE1))
    report = []

    for rec in data.values():
        c1 = (rec.get("collecting_institution_code_1") or "").strip()
        name = (rec.get("collecting_institution") or "").strip()
        if not name:
            continue
        name_norm = normalize(name)

        row = None
        # Step 1: by code
        if c1:
            cur.execute(f"SELECT * FROM {q_lab} WHERE {q_code1}=%s", (c1,))
            row = cur.fetchone()

        # Step 2: by name if not found before
        if row is None:
            cur.execute(
                f"SELECT * FROM {q_lab} WHERE LOWER({_quote('lab_name')})=%s",
                (name_norm,),
            )
            tmp = cur.fetchone()
            # only use the match if your lab_code_1 is empty
            if tmp and not (tmp.get(LAB_COL_CODE1) or "").strip():
                row = tmp

        lab_id = row[LAB_COL_ID] if row else None
        if row and row.get(LAB_COL_CODING):
            taken.add(row[LAB_COL_CODING])

        # ---------- Values to be recorded -------------------------------------
        vals: Dict[str, Any] = {}
        for j, col in FIELD_MAP.items():
            if j in rec:
                v = rec[j]
                if col in ("lab_geo_loc_latitude", "lab_geo_loc_longitude"):
                    v = rounded(v)
                vals[col] = v

        # Fill in lab_code_1 if it is empty in the database.
        if row and c1 and not (row.get(LAB_COL_CODE1) or "").strip():
            vals[LAB_COL_CODE1] = c1

        # apps_name = wetlab if missing (in JSON or DB)
        if not row or not (row.get("apps_name") or "").strip():
            vals.setdefault("apps_name", DEFAULT_APPS)

        vals[LAB_COL_CODING] = unique_code(name, taken)
        for k, d in DEFAULTS.items():
            vals.setdefault(k, d)
        if cid := cities.get(rec.get("geo_loc_city")):
            vals["lab_city_id"] = cid

        # ---------- INSERT / UPDATE -------------------------------------
        if lab_id is None:
            if not a.insert_missing:
                continue
            cols, pars = zip(*vals.items()) if vals else ([], [])
            cur.execute(
                f"INSERT INTO {q_lab} ({', '.join(map(_quote, cols))}) "
                f"VALUES ({', '.join(['%s'] * len(pars))})",
                pars,
            )
            report.append({"action": "insert", "id": cur.lastrowid, "lab": name})
        else:
            if vals:
                setc = ", ".join(f"{_quote(c)}=%s" for c in vals)
                cur.execute(
                    f"UPDATE {q_lab} SET {setc} WHERE {q_id}=%s",
                    list(vals.values()) + [lab_id],
                )
                report.append({"action": "update", "id": lab_id, "lab": name})

    if a.dry_run:
        conn.rollback()
        logging.info("DRY-RUN: %d ops", len(report))
    else:
        conn.commit()
        logging.info("%d applied operations", len(report))
    conn.close()

    if report:
        out = Path(a.report or f"populate_lab_report_{datetime.date.today()}.csv")
        with out.open("w", newline="", encoding="utf-8") as f:
            w = csv.DictWriter(f, fieldnames=sorted(report[0].keys()))
            w.writeheader()
            w.writerows(report)
        logging.info("Report in %s", out)


# ────── CLI ───────────────────────────────────────────────────────────────
if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Sync core_lab_request avoiding duplicates")
    p.add_argument("--lab-json", required=True)
    p.add_argument("--host", default=os.getenv("DB_SERVER_IP", "localhost"))
    p.add_argument("--port", type=int, default=int(os.getenv("DB_PORT", 3306)))
    p.add_argument("--user", default=os.getenv("DB_USER"))
    p.add_argument("--password", default=os.getenv("DB_PASS"))
    p.add_argument("--database", default=os.getenv("DB_NAME", "relecovlims"))
    p.add_argument(
        "--insert-missing",
        action="store_true",
        help="Insert labs that do not exist",
    )
    p.add_argument("--dry-run", action="store_true", help="Rollback at the end")
    p.add_argument("--report", help="Report CSV (optional)")
    main(p.parse_args())



#!/usr/bin/env python
import os
import sys
import argparse
import yaml
import json
from relecov_tools.validate import Validate
from relecov_tools.download import Download
from relecov_tools.config_json import ConfigJson


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-j",
        "--json_file",
        type=str,
        required=True,
        help="Path to the input JSON file containing processed metadata.",
    )
    parser.add_argument(
        "-m",
        "--metadata",
        type=str,
        required=True,
        help="Path to the metadata Excel file.",
    )
    parser.add_argument(
        "--upload_files",
        action="store_true",
        help="If set, upload invalid files to the SFTP server.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=True,
        help="Directory where output files will be written.",
    )
    parser.add_argument(
        "-s",
        "--json_schema_file",
        type=str,
        required=True,
        help="Path to the JSON schema file used for validation.",
    )
    parser.add_argument(
        "-l",
        "--logsum_file",
        type=str,
        required=True,
        help="Path to the previous process's log summary JSON file.",
    )
    args = parser.parse_args()
    print("Creating extra_config yaml")
    conf_file = generate_config_yaml()
    config_json = ConfigJson()
    config_json.include_extra_config(conf_file, config_name=None, force=True)
    print("Fixing filepaths in json file")
    update_json_filepaths(args.json_file)
    print("Initiating validate module")
    validation = Validate(**vars(args))
    validation.user = os.environ["TEST_USER"]
    validation.password = os.environ["TEST_PASSWORD"]
    validation.sftp_port = os.environ["TEST_PORT"]
    validation.execute_validation_process()
    invalid_sftp_folder = validation.remote_outfold.replace("./", "")
    clean_remote_test(invalid_sftp_folder)


def generate_config_yaml():
    """Generate the wrapper_config.yaml file with the desired structure."""
    config_data = {
        "download": {
            "user": "",
            "password": "",
            "conf_file": "",
            "output_dir": "",
            "download_option": "download_clean",
            "subfolder": "RELECOV",
        },
        "validate": {
            "json_file": "",
            "metadata": "",
            "output_dir": "",
            "excel_sheet": "",
            "json_schema_file": "relecov_tools/schema/relecov_schema.json",
        },
    }

    with open("extra_config.yaml", "w") as file:
        yaml.dump(config_data, file, default_flow_style=False)

    return "extra_config.yaml"


def clean_remote_test(invalid_sftp_folder):
    # First clean the repository.
    print("Initating sftp module")

    download = Download(
        user=os.environ["TEST_USER"],
        password=os.environ["TEST_PASSWORD"],
        download_option="download_only",
        output_dir="tests/data/map_validate/",
        subfolder="RELECOV",
    )
    print("Openning connection to sftp")
    download.relecov_sftp.sftp_port = os.environ["TEST_PORT"]
    if not download.relecov_sftp.open_connection():
        print("Could not open connection to remote sftp")
        sys.exit(1)

    print("Cleaning folders inside RELECOV")
    base_paths = ["COD-test-1/RELECOV", "COD-test-2/RELECOV"]
    for base_path in base_paths:
        remote_folders = download.relecov_sftp.list_remote_folders(
            base_path, recursive=True
        )
        remote_folders = sorted(remote_folders, reverse=True)
        for folder in remote_folders:
            folder = folder.replace("./", "")
            if folder == base_path or invalid_sftp_folder not in folder:
                continue

            filelist = download.relecov_sftp.get_file_list(folder)
            for file in filelist:
                print(f"Removing file {file} in {folder}")
                download.relecov_sftp.remove_file(file)

            if "tmp_processing" in folder or "invalid_samples" in folder:
                try:
                    print(f"Removing special folder {folder}")
                    download.relecov_sftp.remove_dir(folder)
                except Exception as e:
                    print(f"Could not remove folder {folder}: {e}")

        filelist_base = download.relecov_sftp.get_file_list(base_path)
        for file in filelist_base:
            print(f"Removing file {file} in {base_path}")
            download.relecov_sftp.remove_file(file)
    return


def update_json_filepaths(json_file):
    with open(json_file, "r") as fh:
        try:
            json_data = json.load(fh)
        except (UnicodeDecodeError, ValueError):
            raise
    updating_fields = ["sequence_file_path_R1", "sequence_file_path_R2"]
    for sample in json_data:
        for field in updating_fields:
            if field not in sample:
                continue
            sample[field] = os.path.join(os.environ["GITHUB_WORKSPACE"], sample[field])
    with open(json_file, "w", encoding="utf-8") as fh:
        fh.write(json.dumps(json_data, indent=4, sort_keys=True, ensure_ascii=False))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from jsonschema import FormatChecker, exceptions
import datetime

import relecov_tools.config_json


def validate_with_exceptions(schema, data, errors):
    """Filter validation errors based on known exceptions.

    This function filters out specific validation errors from the provided list
    of errors based on exceptions defined in the schema. It allows:
      - Numeric fields (integer/float) to contain the placeholder
        'Not Provided [SNOMED:434941000124101]'.
      - String fields with format 'date' to contain the same placeholder.

    Args:
        schema (dict): Dictionary representing the JSON schema.
        data (dict): The input data being validated.
        errors (list): List of validation errors returned by a JSON schema validator.

    Returns:
        filtered_errors (list): List of validation errors excluding known exceptions.
    """
    filtered_errors = []

    for error in errors:
        property_path = ".".join(str(p) for p in error.path)
        prop_schema = schema["properties"].get(property_path, {})

        # allow not provided for numeric types
        if (
            error.validator == "type"
            and error.instance
            == relecov_tools.config_json.ConfigJson().get_topic_data(
                "generic", "not_provided_field"
            )
            and prop_schema.get("type") in ["integer", "number"]
        ):
            continue

        # allow not provided for date format types
        if (
            error.validator == "format"
            and error.instance
            == relecov_tools.config_json.ConfigJson().get_topic_data(
                "generic", "not_provided_field"
            )
            and prop_schema.get("type") == "string"
            and prop_schema.get("format") == "date"
        ):
            continue

        # allow not evaluable for numeric types
        if (
            error.validator == "type"
            and error.instance == "Data Not Evaluable [NCIT:C186292]"
            and prop_schema.get("type") in ["integer", "number"]
        ):
            continue

        # allow not applicable for numeric types
        if (
            error.validator == "type"
            and error.instance == "Not Applicable [GENEPIO:0001619]"
            and prop_schema.get("type") in ["integer", "number"]
        ):
            continue

        # allow not applicable for date format types
        if (
            error.validator == "format"
            and error.instance == "Not Applicable [GENEPIO:0001619]"
            and prop_schema.get("type") == "string"
            and prop_schema.get("format") == "date"
        ):
            continue

        # Keep all other errors
        filtered_errors.append(error)

    return filtered_errors


def make_date_checker(start_date, end_date):
    """Create a format checker to validate dates in the given range only

    Args:
        start_date(datetime.date()): Starting date to set valid range
        end_date(datetime.date()): End date for the valid range

    Returns:
        datechecker (FormatChecker()): Object that will test if properties of
        given format are within the desired range
    """
    checker = FormatChecker()

    @checker.checks("date", raises=exceptions.FormatError)
    def date_range_checker(date):
        if isinstance(date, datetime.datetime):
            date = date.date()
        elif isinstance(date, str):
            try:
                date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                raise exceptions.FormatError(
                    f"'{date}' is not a valid date. Check if it exists or if its format is 'YYYY-MM-DD'"
                )
        elif not isinstance(date, datetime.datetime):
            # Default data type validator will arise this error anyway ('date' is not of type string)
            return True
        if not (start_date <= date <= end_date):
            raise exceptions.FormatError(
                f"Error in date '{date}'. Please provide a date from {start_date} to {end_date}"
            )
        else:
            return True

    return checker



#!/usr/bin/env python
import json
import os
import yaml
import logging

import relecov_tools.utils

log = logging.getLogger(__name__)


# pass test
class ConfigJson:
    """
    ----------------------------------------------------------------------------
    Purpose
    -------
    Load *configuration.json* (defaults) and, optionally, the user
    *extra_config.json* (overrides). Internally we normalise everything to the
    dual-level layout **params / commands**:

        {
          "download": {
              "params": {                    # ← defaults (configuration.json)
                  "threads": 4,
                  "output_dir": "/default/path"
              },
              "commands": {                  # ← overrides (extra_config.json)
                  "threads": 8
              }
          },
          "generic": {...},
          ...
        }

    Helper methods
    --------------
    • ``get_configuration(topic, raw=False)``
        * **raw=False**  → flat dict where *commands overwrite params*
          (ideal to feed modules directly).
        * **raw=True**   → the nested block shown above.

    • ``get_topic_data(topic, key)``
        Returns a *single* value following the priority
        **commands > params > recursive search** >>> legacy flat section.

    ----------------------------------------------------------------------------
    """

    # TODO: Make this path configurable too
    _extra_config_path = os.path.expanduser("~/.relecov_tools/extra_config.json")

    def __init__(
        self,
        json_file=os.path.join(os.path.dirname(__file__), "conf", "configuration.json"),
        extra_config=False,
    ):
        """Load config content in configuration.json and additional config if required

        Parameters
        ----------
        json_file : str
            Path to the *defaults* file (``configuration.json``).
        extra_config : bool
            If *True* merge in the user's overrides located at
            ``~/.relecov_tools/extra_config.json``.
        """
        # ── 1. Load defaults ------------------------------------------------
        with open(json_file, "r", encoding="utf-8") as fh:
            base_conf = json.load(fh)

        # ── 2. Optionally load user overrides -------------------------------
        extra_conf, active_extra = {}, False
        if extra_config and os.path.isfile(ConfigJson._extra_config_path):
            try:
                with open(ConfigJson._extra_config_path, "r", encoding="utf-8") as fh:
                    extra_conf = json.load(fh)
                active_extra = True
            except (OSError, json.JSONDecodeError) as e:
                log.warning(f"Could not load extra config: {e}. Using default instead")
        elif extra_config:
            log.warning(
                f"Could not load extra config: {ConfigJson._extra_config_path} does not exist. "
                "Using default instead"
            )
            log.warning(
                "Run `relecov-tools add-extra-config` to include additional configuration"
            )

        # ── 3. Build an index  <leaf-key → first-level-parent> ---------------
        #    Needed to relocate overrides that appear outside their branch
        self._leaf_parent = {}

        def _index_parents(node: dict, top_key: str):
            if not isinstance(node, dict):
                return
            for k, v in node.items():
                self._leaf_parent.setdefault(k, top_key)
                _index_parents(v, top_key)

        for first_level_key, subtree in base_conf.items():
            _index_parents(subtree, first_level_key)

        # ── 4. Merge defaults + overrides into params/commands ---------------
        self.json_data = self._nested_merge_with_commands(base_conf, extra_conf)

        log.debug(
            "Loaded additional configuration."
            if active_extra
            else "Running with default configuration."
        )
        self.topic_config = list(self.json_data.keys())

    def get_configuration(self, topic: str, *, raw: bool = False):
        """
        Return the configuration block for *topic*.

        Examples
        --------
        >>> cfg = ConfigJson(extra_config=True)
        >>> cfg.get_configuration("download")
        {'threads': 8, 'output_dir': '/default/path'}          # flat view

        >>> cfg.get_configuration("download", raw=True)
        {'params': {'threads': 4, 'output_dir': '/default/path'},
         'commands': {'threads': 8}}                          # nested view
        """
        # ── 1. Topic not present ───────────────────────────────────────────
        if topic not in self.json_data:
            return None

        block = self.json_data[topic]

        # ── 2. Caller wants the nested structure as is ─────────────────────
        if raw:
            return block

        # ── 3. Layout: merge params + commands  (commands win) ─────────
        if isinstance(block, dict) and ("params" in block or "commands" in block):
            flattened = dict(block.get("params", {}))  # defaults
            flattened.update(block.get("commands", {}))  # overrides
            return flattened

        return block

    def get_topic_data(self, topic: str, found: str):
        """
        Fetch a single value *found* from *topic*.

        Search priority
        ----------------
        1. **commands** – overrides
        2. **params**   – defaults
        3. Recursive search inside both dicts
        4. Legacy flat section (back-compat)

        Example
        -------
        >>> cfg.get_topic_data("download", "threads")   # → 8
        >>> cfg.get_topic_data("download", "output_dir")  # → '/default/path'
        """

        # ── Helper: depth-first search in nested dicts ──────────────────────
        def _recursive_lookup(node, key):
            """Searches for `key` at any level of nested dictionaries."""
            if not isinstance(node, dict):
                return None
            for k, v in node.items():
                if k == key:
                    return v
                if isinstance(v, dict):
                    res = _recursive_lookup(v, key)
                    if res is not None:
                        return res
            return None

        # ── 1. Find the topic block ─────────────────────────────────────────
        topic_block = self.json_data.get(topic)
        if topic_block is None:
            return None

        # ── 2. New layout (params / commands) ───────────────────────────────
        if isinstance(topic_block, dict) and (
            "params" in topic_block or "commands" in topic_block
        ):
            # 2.1  direct hit in commands  (highest priority)
            if found in topic_block.get("commands", {}):
                return topic_block["commands"][found]
            # 2.2  direct hit in params   (defaults)
            if found in topic_block.get("params", {}):
                return topic_block["params"][found]
            # 2.3  recursive search (first commands, then params)
            return _recursive_lookup(
                topic_block.get("commands", {}), found
            ) or _recursive_lookup(topic_block.get("params", {}), found)

        # ── 3. Legacy flat section ──────────────────────────────────────────
        if found in topic_block:
            return topic_block[found]

        # ── 4. Legacy with deeper nesting ───────────────────────────────────
        return _recursive_lookup(topic_block, found)

    def include_extra_config(self, config_file, config_name=None, force=False):
        """Include given file content as additional configuration for later usage.

        Args:
            config_name (str): Name of the config key that will be added.
            config_file (str): Path to the input file, either Json or Yaml format.
            force (bool, optional): Force replacement of existing configuration if needed.

        Raises:
            ValueError: If provided config_file does not have a supported extension.
        """

        def validate_new_config(config_name, current_conf, force=False):
            """Check if config_name is already in configuration"""
            if config_name in current_conf.keys():
                if force:
                    log.info(
                        f"`{config_name}` already in config. Replacing its content..."
                    )
                    return True
                else:
                    err_txt = f"Cannot add `{config_name}`: already in config. Set `force` to force replacement"
                    log.error(err_txt)
                    return False
            else:
                return True

        def rec_merge_config(current_conf, new_conf, force=False):
            """Recursively add new configuration without deleting the existing keys"""
            for k, v in new_conf.items():
                if isinstance(v, dict) and k in current_conf:
                    rec_merge_config(current_conf[k], v, force=force)
                else:
                    change_msg = f"{k}: {current_conf.get(k, '')} -> {v}"
                    if not validate_new_config(k, current_conf, force=force):
                        summary["Canceled"].append(change_msg)
                        continue
                    summary["Included"].append(change_msg)
                    current_conf[k] = v

        if not os.path.isfile(str(config_file)):
            raise FileNotFoundError(f"Extra config file {config_file} does not exist")
        if os.path.isfile(ConfigJson._extra_config_path):
            with open(ConfigJson._extra_config_path, "r", encoding="utf-8") as f:
                additional_config = json.load(f)
        else:
            additional_config = None

        summary = {"Included": [], "Canceled": []}
        valid_exts = (".json", ".yaml", ".yml")
        os.makedirs(os.path.dirname(ConfigJson._extra_config_path), exist_ok=True)
        with open(config_file, "r") as fh:
            if config_file.endswith(".json"):
                file_content = json.load(fh)
            elif config_file.endswith((".yaml", ".yml")):
                file_content = yaml.load(fh, Loader=yaml.FullLoader)
            else:
                raise ValueError(
                    f"config_file {config_file} extension is not supported. Use: {valid_exts}"
                )
        if additional_config is not None:
            if config_name is None:
                rec_merge_config(additional_config, file_content, force=force)
            else:
                if validate_new_config(config_name, self.json_data, force=force):
                    msg = f"{config_name}: {file_content}"
                    summary["Included"].append(msg)
                    additional_config.update({config_name: file_content})
        else:
            if config_name is None:
                additional_config = file_content
            else:
                additional_config = {config_name: file_content}
            summary["Included"].extend(list(additional_config.keys()))
        relecov_tools.utils.write_json_to_file(
            additional_config, ConfigJson._extra_config_path
        )
        log.info("Finished including extra configuration")
        print("Update summary:")
        for state, changes in summary.items():
            print(state, ":\n", "\n".join([str(msg) for msg in changes]))
        return

    def remove_extra_config(self, config_name):
        """Remove key from extra_config configuration file

        Args:
            config_name (str): _description_
        """
        if config_name is None:
            try:
                os.remove(ConfigJson._extra_config_path)
                log.info("Removed extra config file.")
            except OSError as e:
                log.error(f"Could not remove extra config file: {e}")
                return
        else:
            with open(ConfigJson._extra_config_path, "r") as fh:
                additional_config = json.load(fh)
            try:
                additional_config.pop(config_name)
            except KeyError:
                log.error(f"{config_name} not found in extra_config for removal")
                return
            except OSError as e:
                log.error(f"Could not remove {config_name} key: {e}")
                return
            relecov_tools.utils.write_json_to_file(
                additional_config, ConfigJson._extra_config_path
            )
            log.info(f"Successfully removed {config_name} from extra config")
        print(f"Finished clearing extra config: {config_name}")
        return

    def get_lab_code(self, submitting_institution):
        """Get the corresponding code for the given submitting institution"""
        if submitting_institution is None:
            log.warning("No submitting institution could be found to update lab_code")
            return
        institutions_config = self.get_configuration("institutions_config")
        if not institutions_config:
            log.warning("No institutions_config found. Could not extract lab_code")
            return
        for code, conf in institutions_config.items():
            if submitting_institution in conf.get("institution_name"):
                return code
        else:
            log.warning(f"{submitting_institution} not found in institutions_config")
        return

    def _nested_merge_with_commands(self, base_conf: dict, extra_conf: dict) -> dict:
        """
        Produce the *params / commands* structure described in the class docstring.

        • Everything from *configuration.json* → **params**
        • Everything from *extra_config.json* → **commands**
          (relocating the key to its first-level parent if necessary).
        """
        merged = {}

        for key, val in base_conf.items():
            merged[key] = {"params": val, "commands": {}}

        for key, val in extra_conf.items():
            if key in merged:
                merged[key]["commands"] = val
            elif key in self._leaf_parent:
                parent = self._leaf_parent[key]
                merged.setdefault(parent, {"params": {}, "commands": {}})
                merged[parent]["commands"][key] = val
            else:
                merged[key] = {"params": {}, "commands": val}

        return merged



#!/usr/bin/env python
import json
import rich.console
import os
import re
from datetime import datetime as dtime
import relecov_tools.utils
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class LabMetadata(BaseModule):
    def __init__(
        self,
        metadata_file=None,
        sample_list_file=None,
        output_dir=None,
        files_folder=None,
        **kwargs,
    ):
        super().__init__(output_dir=output_dir, called_module=__name__)
        self.log.info("Initiating read-lab-metadata process")
        self.sample_list_file = sample_list_file
        self.files_folder = files_folder

        if metadata_file is None:
            self.metadata_file = relecov_tools.utils.prompt_path(
                msg="Select the excel file which contains metadata"
            )
        else:
            self.metadata_file = metadata_file

        if not os.path.exists(self.metadata_file):
            self.log.error("Metadata file %s does not exist ", self.metadata_file)
            stderr.print(
                "[red] Metadata file " + self.metadata_file + " does not exist"
            )
            raise FileNotFoundError(f"Metadata file {self.metadata_file} not found")

        if sample_list_file is None:
            stderr.print("[yellow]No samples_data.json file provided")
            self.log.warning("No samples_data.json file provided")
            if not os.path.isdir(str(files_folder)):
                stderr.print("[red]No samples file nor valid files folder provided")
                self.log.error("No samples file nor valid files folder provided")
                raise FileNotFoundError(
                    "No samples file nor valid files folder provided"
                )
            self.files_folder = os.path.abspath(files_folder)

        if sample_list_file is not None and not os.path.exists(sample_list_file):
            self.log.error(
                "Sample information file %s does not exist ", sample_list_file
            )
            stderr.print("[red] Samples file " + sample_list_file + " does not exist")
            raise FileNotFoundError(
                "Sample information file %s does not exist ", sample_list_file
            )
        else:
            self.sample_list_file = sample_list_file

        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = output_dir

        self.config_json = ConfigJson(extra_config=True)

        # TODO: remove hardcoded schema selection
        relecov_schema = self.config_json.get_topic_data("generic", "relecov_schema")
        relecov_sch_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "schema", relecov_schema
        )
        self.configuration = self.config_json
        self.institution_config = self.config_json.get_configuration(
            "institutions_config"
        )

        out_path = os.path.realpath(self.output_dir)
        self.lab_code = out_path.split("/")[-2]
        self.logsum = self.parent_log_summary(
            output_dir=self.output_dir, lab_code=self.lab_code, path=out_path
        )

        with open(relecov_sch_path, "r") as fh:
            self.relecov_sch_json = json.load(fh)

        try:
            relecov_tools.assets.schema_utils.jsonschema_draft.check_schema_draft(
                self.relecov_sch_json, "2020-12"
            )
        except Exception as e:
            self.log.error("JSON schema is not valid: %s", str(e))
            stderr.print(f"[red]Error: JSON schema is not valid.\n{str(e)}")
            raise

        self.label_prop_dict = {}

        for prop, values in self.relecov_sch_json["properties"].items():
            try:
                self.label_prop_dict[values["label"]] = prop
            except KeyError:
                self.log.warning("Property %s does not have 'label' attribute", prop)
                stderr.print(
                    "[orange]Property " + prop + " does not have 'label' attribute"
                )
                continue
        self.date = dtime.now().strftime("%Y%m%d%H%M%S")
        self.json_req_files = self.config_json.get_topic_data(
            "read_lab_metadata", "lab_metadata_req_json"
        )
        self.schema_name = self.relecov_sch_json["title"]
        self.schema_version = self.relecov_sch_json["version"]
        self.metadata_processing = self.config_json.get_topic_data(
            "sftp_handle", "metadata_processing"
        )
        self.samples_json_fields = self.config_json.get_topic_data(
            "read_lab_metadata", "samples_json_fields"
        )
        self.unique_sample_id = "sequencing_sample_id"

    def _split_institution(self, raw: str) -> tuple[str, str]:
        """
        Split a free-text institution string into:

            1. **name** – the visible name without any bracketed tags.
            2. **code** – the last bracketed element, assumed to be the CCN
            (e.g. `[1328000027]`). If no second bracket exists, returns "".

        Example
        -------
        _split_institution("Hospital X [Madrid] [1328000027]")
        ("Hospital X", "1328000027")
        """
        name = re.split(r"\s*\[", raw, maxsplit=1)[0].strip()
        brackets = re.findall(r"\[([^\]]+)\]", raw)
        code = brackets[-1].strip() if len(brackets) >= 2 else ""
        return name, code

    INSTITUTION_FIELDS = {
        "collecting_institution",
        "submitting_institution",
        "sequencing_institution",
    }

    def get_samples_files_data(self, clean_metadata_rows):
        """Include the fields that would be included in samples_data.json

        Args:
            clean_metadata_rows (list(dict)): Cleaned list of rows from metadata_lab.xlsx file

        Returns:
            j_data (dict(dict)): Dictionary where each key is the sample ID and the values are
            its file names, locations and md5
        """

        def safely_calculate_md5(file):
            """Check file md5, but return Not Provided if file does not exist"""
            self.log.info("Generating md5 hash for %s...", str(file))
            try:
                return relecov_tools.utils.calculate_md5(file)
            except IOError:
                return self.config_json.get_topic_data("generic", "not_provided_field")

        # The files are and md5file are supposed to be located together
        dir_path = self.files_folder
        md5_checksum_files = [
            os.path.join(dir_path, f) for f in os.listdir(dir_path) if "md5" in f
        ]
        if md5_checksum_files:
            skip_list = self.configuration.get_topic_data(
                "sftp_handle", "skip_when_found"
            )
            md5_dict = relecov_tools.utils.read_md5_checksum(
                file_name=md5_checksum_files[0], avoid_chars=skip_list
            )
        else:
            md5_dict = {}
            self.log.warning("No md5sum file found.")
            self.log.warning("Generating new md5 hashes. This might take a while...")
        j_data = {}
        no_fastq_error = "No R1 fastq file was given for sample %s in metadata"
        n = 0
        for sample in clean_metadata_rows:
            n += 1
            sample_id = str(sample.get(self.unique_sample_id))
            if not sample_id:
                if sample.get("collecting_lab_sample_id"):
                    sample_id = sample["collecting_lab_sample_id"]
                    self.unique_sample_id = "collecting_lab_sample_id"
                else:
                    sample_id = sample.get("sequence_file_R1", "").split(".")[0]
                    self.unique_sample_id = "sequence_file_R1"
            files_dict = {}
            r1_file = sample.get("sequence_file_R1")
            r2_file = sample.get("sequence_file_R2")
            if not r1_file:
                self.logsum.add_error(
                    sample=sample_id,
                    entry=no_fastq_error % sample_id,
                )
                j_data[sample_id] = files_dict
                continue
            r1_md5 = md5_dict.get(r1_file)
            r2_md5 = md5_dict.get(r2_file)
            files_dict["sequence_file_R1"] = r1_file
            files_dict["sequence_file_path_R1"] = dir_path
            batch_id = dir_path.split("/")[-1]
            logtxt = f"Setting batch_id to {batch_id} based on download dir: {dir_path}"
            stderr.print(f"[yellow]{logtxt}")
            self.log.info(logtxt)
            files_dict["batch_id"] = dir_path.split("/")[-1]
            if not os.path.exists(os.path.join(dir_path, r1_file)):
                self.logsum.add_error(
                    sample=sample_id, entry="Provided R1 file not found after download"
                )
                continue
            if r1_md5:
                files_dict["sequence_file_R1_md5"] = r1_md5
            else:
                files_dict["sequence_file_R1_md5"] = safely_calculate_md5(
                    os.path.join(dir_path, r1_file)
                )
            if r2_file:
                files_dict["sequence_file_R2"] = r2_file
                files_dict["sequence_file_path_R2"] = dir_path
                if not os.path.exists(os.path.join(dir_path, r2_file)):
                    self.logsum.add_error(
                        sample=sample_id,
                        entry="Provided R2 file not found after download",
                    )
                    continue
                if r2_md5:
                    files_dict["sequence_file_R2_md5"] = r2_md5
                else:
                    files_dict["sequence_file_R2_md5"] = safely_calculate_md5(
                        os.path.join(dir_path, r1_file)
                    )
            if sample_id in j_data:
                sample_id = "_".join([sample_id, str(n)])
            j_data[sample_id] = files_dict
        if not any(val for val in j_data.values()):
            errtxt = f"No files found for the samples in {dir_path}"
            self.logsum.add_error(entry=errtxt, sample=sample_id)
            stderr.print(f"[red]{errtxt}")
        try:
            samples_filename = "_".join(["samples_data", self.lab_code + ".json"])
            samples_filename = self.tag_filename(filename=samples_filename)
            file_path = os.path.join(self.output_dir, samples_filename)
            relecov_tools.utils.write_json_to_file(j_data, file_path)
        except Exception:
            self.log.error("Could not output samples_data.json file to output folder")
        return j_data

    def match_to_json(self, valid_metadata_rows):
        """Keep only the rows from samples present in the input file samples.json

        Args:
            valid_metadata_rows (list(dict)): List of rows from metadata_lab.xlsx file

        Returns:
            clean_metadata_rows(list(dict)): List of rows matching the samples in samples_data.json
            missing_samples(list(str)): List of samples not found in samples_data.json
        """
        missing_samples = []
        if not self.sample_list_file:
            logtxt = "samples_data.json not provided, all samples will be included"
            self.logsum.add_warning(entry=logtxt)
            return valid_metadata_rows, missing_samples
        else:
            samples_json = relecov_tools.utils.read_json_file(self.sample_list_file)
        clean_metadata_rows = []
        for row in valid_metadata_rows:
            sample_id = str(row[self.unique_sample_id]).strip()
            self.logsum.feed_key(sample=sample_id)
            if sample_id in samples_json.keys():
                clean_metadata_rows.append(row)
            else:
                log_text = "Sample in metadata but missing in downloaded samples file"
                self.logsum.add_warning(sample=sample_id, entry=log_text)
                missing_samples.append(sample_id)
        return clean_metadata_rows, missing_samples

    def adding_fixed_fields(self, m_data):
        """Include fixed data that are always the same for every sample"""
        p_data = self.configuration.get_topic_data("read_lab_metadata", "fixed_fields")
        organism_mapping = self.configuration.get_topic_data(
            "read_lab_metadata", "organism_mapping"
        )

        for idx in range(len(m_data)):
            organism = m_data[idx].get("organism", "")
            if organism in organism_mapping:
                m_data[idx]["tax_id"] = organism_mapping[organism]["tax_id"]
                m_data[idx]["host_disease"] = organism_mapping[organism]["host_disease"]
            else:
                m_data[idx]["tax_id"] = "Missing [LOINC:LA14698-7]"
                m_data[idx]["host_disease"] = "Missing [LOINC:LA14698-7]"
            for key, value in p_data.items():
                m_data[idx][key] = value
            m_data[idx]["schema_name"] = self.schema_name
            m_data[idx]["schema_version"] = self.schema_version
            m_data[idx]["submitting_institution_id"] = self.lab_code
        return m_data

    def adding_copy_from_other_field(self, m_data):
        """Add a new field with information based in another field."""
        p_data = self.configuration.get_topic_data(
            "read_lab_metadata", "required_copy_from_other_field"
        )
        for idx in range(len(m_data)):
            for key, value in p_data.items():
                m_data[idx][key] = m_data[idx][value]
        return m_data

    def adding_post_processing(self, m_data):
        """Add fields which values require post processing"""
        p_data = self.configuration.get_topic_data(
            "read_lab_metadata", "required_post_processing"
        )
        for idx in range(len(m_data)):
            for key, p_values in p_data.items():
                value = m_data[idx].get(key)
                if not value:
                    continue
                if value in p_values:
                    p_field, p_set = p_values[value].split("::")
                    m_data[idx][p_field] = p_set
                else:
                    # Check if key p_values should match only part of the value
                    for reg_key, reg_value in p_values.items():
                        if reg_key in value:
                            p_field, p_set = reg_value.split("::")
                            m_data[idx][p_field] = p_set

        return m_data

    def adding_ontology_to_enum(self, m_data):
        """Read the schema to get the properties enum and, for those fields
        which have an enum property value, replace the value for the one
        that is defined in the schema.
        """
        enum_dict = {}
        for prop, values in self.relecov_sch_json["properties"].items():
            enum_values = values.get("enum", [])
            ontologies_present = any(
                isinstance(enum, str) and re.search(r" \[\w+:.*\]$", enum)
                for enum in enum_values
            )
            if not ontologies_present:
                continue
            if "enum" in values:
                enum_dict[prop] = {}
                for enum in values["enum"]:
                    go_match = re.search(r"(.+) \[\w+:.*", enum)
                    if go_match:
                        enum_dict[prop][go_match.group(1)] = enum
                    else:
                        enum_dict[prop][enum] = enum
        ontology_errors = {}
        for idx in range(len(m_data)):
            for key, e_values in enum_dict.items():
                if key in m_data[idx]:
                    current_value = m_data[idx][key]
                    if re.search(r" \[\w+:.*\]$", current_value):
                        continue  # If already has ontology, do nothing.
                    if current_value in e_values:
                        m_data[idx][key] = e_values[current_value]
                    else:
                        sample_id = m_data[idx][self.unique_sample_id]
                        log_text = f"No ontology found for {current_value} in {key}"
                        self.logsum.add_warning(sample=sample_id, entry=log_text)
                        ontology_errors[key] = ontology_errors.get(key, 0) + 1
                        continue
        if len(ontology_errors) >= 1:
            stderr.print(
                "[red] No ontology could be added in:\n",
                "\n".join({f"{x} - {y} samples" for x, y in ontology_errors.items()}),
            )
            self.log.warning(
                "No ontology could be added in:\n%s",
                "\n".join(f"{x} - {y} samples" for x, y in ontology_errors.items()),
            )

        return m_data

    def process_from_json(self, m_data, json_fields):
        """
        Fill in the fields defined in *json_fields* for each sample.

        This is done in two phases:

        1.  **Per-sample
            - If *map_field* is empty → warning in `log_summary.json`.
            - If the code exists but is not found in the auxiliary JSON → warning in `log_summary.json`.
            - If the code exists and is found → add the fields specified in *adding_fields*.
            - If any of these fields are `Not Provided`, they are filled with `Not Provided [SNOMED:434941000124101]`.

        2.  **console summary** (`stderr`)
            - On completion, a **single line** per type of problem
              (empty codes or unknown codes) is displayed indicating how many samples
              are affected and the affected property.
            Example:
                `13 samples without CCN; check log`.

        Parameters
        ----------
            m_data : list[dict]
            Metadata already read from Excel.
            json_fields : dict
            Dict taken from *configuration.json* with:
                        - map_field
                        - adding_fields
                        - file
                        - j_data (previously loaded in adding_fields())

        Returns
        -------
            list[dict]
            Metadata with the new fields added.
        """
        map_field = json_fields["map_field"]
        col_label = self.relecov_sch_json["properties"][map_field]["label"]
        json_data = json_fields["j_data"]

        # ─── counters for the summary ─────────────────────────────────────────
        empty_codes = []  # samples without value in map_field
        unknown_codes = []  # value present but not found in auxiliary json

        for row in m_data:
            sample_id = str(row.get(self.unique_sample_id))
            code = (row.get(map_field) or "").strip()

            # ╭─ 1. Empty field ───────────────────────────────────────────────╮
            if not code:
                msg = (
                    f"{col_label} not provided; cannot map "
                    f"{json_fields['file']} data"
                )
                self.logsum.add_warning(sample=sample_id, entry=msg)
                empty_codes.append(sample_id)
                continue
            # ╰────────────────────────────────────────────────────────────────╯

            # ─── 2. Attempt Mapping ────────────────────────────────────────────
            try:
                adding_data = {
                    k: v
                    for k, v in json_data[code].items()
                    if k in json_fields["adding_fields"]
                }
                row.update(adding_data)

            except KeyError:
                # Code present but does not exist in the auxiliary JSON
                msg = (
                    f"Unknown {col_label} '{code}' in {json_fields['file']} "
                    f"for sample {sample_id}"
                )
                self.logsum.add_warning(sample=sample_id, entry=msg)
                unknown_codes.append(sample_id)
                continue

            # ─── 3. Fill Not Provided in the added fields ─────────────
            for field in json_fields["adding_fields"]:
                if (
                    field not in row
                    or not row[field]
                    or str(row[field]).lower().startswith("not provided")
                ):
                    row[field] = self.config_json.get_topic_data(
                        "generic", "not_provided_field"
                    )

        # ─── 4. Summary in stderr ──────────────────────────────────────────
        if empty_codes:
            stderr.print(
                f"[yellow]{len(empty_codes)} samples without {col_label}; "
                "check the log"
            )
        if unknown_codes:
            stderr.print(
                f"[yellow]{len(unknown_codes)} values of {col_label} not found "
                f"en {json_fields['file']}; check the log"
            )

        return m_data

    def infer_file_format_from_schema(self, metadata):
        """Infer the file_format field based on the extension in sequence_file_R1,
        using enum values (with ontology) directly from the schema."""

        extension_map = {
            ".fastq": "FASTQ",
            ".fastq.gz": "FASTQ",
            ".fq": "FASTQ",
            ".fq.gz": "FASTQ",
            ".bam": "BAM",
            ".cram": "CRAM",
            ".fasta": "FASTA",
            ".fa": "FASTA",
        }

        file_format_enum = (
            self.relecov_sch_json["properties"].get("file_format", {}).get("enum", [])
        )
        keyword_to_enum = {}

        for item in file_format_enum:
            match = re.match(r"^([A-Z]+)", item)
            if match:
                keyword = match.group(1)
                keyword_to_enum[keyword] = item

        for row in metadata:
            r1_file = row.get("sequence_file_R1", "").lower()
            file_format_val = None
            for ext, keyword in extension_map.items():
                if r1_file.endswith(ext.lower()):
                    file_format_val = keyword_to_enum.get(keyword)
                    break

            if file_format_val:
                row["file_format"] = file_format_val
            else:
                row["file_format"] = self.config_json.get_topic_data(
                    "generic", "not_provided_field"
                )

        return metadata

    def adding_fields(self, metadata):
        """Add information located inside various json file as fields"""

        for key, values in self.json_req_files.items():
            stderr.print(f"[blue]Processing {key}")
            self.log.info(f"Processing {key}")
            f_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "conf", values["file"]
            )
            values["j_data"] = relecov_tools.utils.read_json_file(f_path)
            metadata = self.process_from_json(metadata, values)
            stderr.print(f"[green]Processed {key}")
            self.log.info(f"Processed {key}")

        if self.institution_config:
            self.log.info("Updating laboratory code from institutions_config...")
        # Include Sample information data from sample json file
        stderr.print("[blue]Processing sample data file")
        self.log.info("Processing sample data file")
        s_json = {}
        # TODO: Change sequencing_sample_id for some unique ID used in RELECOV database
        s_json["map_field"] = self.unique_sample_id
        s_json["adding_fields"] = self.samples_json_fields
        if self.sample_list_file:
            s_json["j_data"] = relecov_tools.utils.read_json_file(self.sample_list_file)
        else:
            s_json["j_data"] = self.get_samples_files_data(metadata)
        if not s_json["j_data"]:
            self.log.warning(
                f"Samples file {self.sample_list_file} is empty. All samples will be included"
            )
            s_json["j_data"] = self.get_samples_files_data(metadata)
        batch_id = self.get_batch_id_from_data(list(s_json["j_data"].values()))
        # This will declare self.batch_id in BaseModule() which will be used later
        self.set_batch_id(batch_id)
        metadata = self.process_from_json(metadata, s_json)
        metadata = self.infer_file_format_from_schema(metadata)
        stderr.print("[green]Processed sample data file.")
        self.log.info("Processed sample data file.")
        return metadata

    def read_configuration_json_files(self):
        """Read json files defined in configuration lab_metadata_req_json
        property
        """
        c_files = {}
        for item, value in self.json_files.items():
            f_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "conf", value
            )
            c_files[item] = relecov_tools.utils.read_json_file(f_path)
        return c_files

    def read_metadata_file(self):
        """
        Reads the input metadata file, converts headings to schema properties
        and returns a list[dict] with the cleaned rows.
        """
        meta_sheet = self.metadata_processing.get("excel_sheet")
        header_flag = self.metadata_processing.get("header_flag")
        sample_id_col = self.metadata_processing.get("sample_id_col")
        self.alternative_heading = False

        # ────────── 0. Open sheet ──────────
        try:
            ws_metadata_lab, heading_row_number = relecov_tools.utils.read_excel_file(
                self.metadata_file, meta_sheet, header_flag, leave_empty=False
            )
        except KeyError:
            self.alternative_heading = True
            alt_sheet = self.metadata_processing.get("alternative_sheet")
            header_flag = self.metadata_processing.get("alternative_flag")
            sample_id_col = self.metadata_processing.get("alternative_sample_id_col")
            logtxt = f"No excel sheet named {meta_sheet}. Using {alt_sheet}"
            stderr.print(f"[yellow]{logtxt}")
            self.log.error(logtxt)
            ws_metadata_lab, heading_row_number = relecov_tools.utils.read_excel_file(
                self.metadata_file, alt_sheet, header_flag, leave_empty=False
            )

        valid_metadata_rows, included_sample_ids = [], []
        row_number = heading_row_number

        for row in ws_metadata_lab:
            row_number += 1
            property_row = {}

            # ────────── 1. Validate sample_id ──────────
            try:
                sample_id = str(row[sample_id_col]).strip()
            except KeyError:
                self.logsum.add_error(entry=f"No {sample_id_col} found in excel file")
                continue

            if sample_id in included_sample_ids:
                log_text = (
                    f"Skipped duplicated sample {sample_id} in row {row_number}. "
                    f"Sequencing sample id must be unique"
                )
                self.logsum.add_warning(entry=log_text)
                continue

            if not row[sample_id_col] or "Not Provided" in sample_id:
                if row.get("collecting_lab_sample_id"):
                    sample_id = row["collecting_lab_sample_id"]
                else:
                    sample_id = row.get("sequence_file_R1", "").split(".")[0]
                if not sample_id:
                    log_text = (
                        f"{sample_id_col} not provided in row {row_number}. Skipped"
                    )
                    self.logsum.add_error(entry=log_text)
                    stderr.print(f"[red]{log_text}")
                    continue
                else:
                    log_text = f"{sample_id_col} not provided for {sample_id}"
                    self.logsum.add_error(entry=log_text, sample=sample_id)
                    stderr.print(f"[red]{log_text}")

            included_sample_ids.append(sample_id)

            # ────────── 2. Process columns ──────────
            for key in row.keys():
                if header_flag in key:
                    continue

                schema_key = self.label_prop_dict.get(key, key)
                value = row[key]

                # 2.1  Institution fields
                if schema_key in self.INSTITUTION_FIELDS and value:
                    name, code = self._split_institution(str(value))
                    value = name
                    if schema_key == "collecting_institution":
                        if code:
                            property_row["collecting_institution_code_1"] = code
                        else:
                            self.logsum.add_warning(
                                sample=sample_id,
                                entry="CCN not provided for collecting_institution",
                            )

                # 2.2  Omit empty properties
                if value is None or value == "" or "not provided" in str(value).lower():
                    log_text = f"{key} not provided for sample {sample_id}"
                    self.logsum.add_warning(sample=sample_id, entry=log_text)
                    continue

                # 2.3  Cast to schema type
                schema_type = (
                    self.relecov_sch_json["properties"]
                    .get(schema_key, {})
                    .get("type", "string")
                )
                try:
                    value = relecov_tools.utils.cast_value_to_schema_type(
                        value, schema_type
                    )
                except (ValueError, TypeError) as e:
                    log_text = (
                        f"Type conversion error for {key} (expected {schema_type}): "
                        f"{value}. {e}"
                    )
                    self.logsum.add_error(sample=sample_id, entry=log_text)
                    stderr.print(f"[red]{log_text}")
                    continue

                # 2.4  Normalise dates and numbers
                if "date" in key.lower():
                    pattern = r"^\d{4}[-/.]\d{2}[-/.]\d{2}"
                    if isinstance(row[key], dtime):
                        value = str(row[key].date())
                    elif re.match(pattern, str(row[key])):
                        value = re.match(
                            pattern, str(row[key]).replace("/", "-").replace(".", "-")
                        ).group(0)
                    else:
                        try:
                            value = str(int(float(str(row[key]))))
                            self.log.info(
                                "Date given as an integer. Understood as a year"
                            )
                        except (ValueError, TypeError):
                            log_text = f"Invalid date format in {key}: {row[key]}"
                            self.logsum.add_error(sample=sample_id, entry=log_text)
                            stderr.print(f"[red]{log_text} for sample {sample_id}")
                            continue
                elif "sample id" in key.lower() and isinstance(row[key], (float, int)):
                    value = str(int(row[key]))
                elif isinstance(row[key], (float, int)):
                    value = str(row[key])

                if "date" not in key.lower() and isinstance(row[key], dtime):
                    logtxt = f"Non-date field {key} provided as date. Parsed as int"
                    self.logsum.add_warning(sample=sample_id, entry=logtxt)
                    value = str(relecov_tools.utils.excel_date_to_num(row[key]))

                property_row[schema_key] = value

            valid_metadata_rows.append(property_row)

        return valid_metadata_rows

    def create_metadata_json(self):
        stderr.print("[blue]Reading Lab Metadata Excel File")
        valid_metadata_rows = self.read_metadata_file()
        stderr.print(f"[green]Processed {len(valid_metadata_rows)} valid metadata rows")
        self.log.info(f"Processed {len(valid_metadata_rows)} valid metadata rows")
        clean_metadata_rows, missing_samples = self.match_to_json(valid_metadata_rows)
        if missing_samples:
            num_miss = len(missing_samples)
            logtx = "%s samples not found in metadata: %s" % (num_miss, missing_samples)
            self.logsum.add_warning(entry=logtx)
            stderr.print(f"[yellow]{num_miss} samples missing:\n{missing_samples}")
        # Continue by adding extra information
        stderr.print("[blue]Including additional information")
        self.log.info("Including additional information")

        extended_metadata = self.adding_fields(clean_metadata_rows)
        stderr.print("[blue]Including post processing information")
        self.log.info("Including post processing information")
        extended_metadata = self.adding_post_processing(extended_metadata)
        extended_metadata = self.adding_copy_from_other_field(extended_metadata)
        extended_metadata = self.adding_fixed_fields(extended_metadata)
        completed_metadata = self.adding_ontology_to_enum(extended_metadata)
        if not completed_metadata:
            self.log.warning("Metadata was completely empty. No output file generated")
            stderr.print("Metadata was completely empty. No output file generated")
            return
        file_code = "_".join(["read_lab_metadata", self.lab_code]) + ".json"
        file_name = self.tag_filename(filename=file_code)
        stderr.print("[blue]Writting output json file")
        os.makedirs(self.output_dir, exist_ok=True)
        # Creating log summary
        self.parent_create_error_summary()
        file_path = os.path.join(self.output_dir, file_name)
        self.log.info("Writting output json file %s", file_path)
        relecov_tools.utils.write_json_to_file(completed_metadata, file_path)
        return True
